import os
from typing import Optional, Dict, Any, Tuple, List

import h5py
import json

from . import path_utils

_DATA_FORMATS = ("esrfv1", "esrfv2", "esrfv3", "id16bspec", "mx")

_BEAMLINE_DIR_TO_NAME = {
    "id30a1": "id30a-1",
    "id30a3": "id30a-3",
    "id23eh1": "id23-1",
    "id23eh2": "id23-2",
}
_BEAMLINE_NAME_TO_DIR = {
    "id30a-1": "id30a1",
    "id30a-3": "id30a3",
    "id23-1": "id23eh1",
    "id23-2": "id23eh2",
}

MX_METADATA_FILENAME = "metadata.json"


def get_session_dir(
    proposal: str,
    beamline: str,
    session: str,
    root_dir: Optional[str] = None,
    raw_data_format: str = "esrfv3",
) -> str:
    """Get the session directory from the proposal, beamlines and session name."""
    if raw_data_format in _DATA_FORMATS:
        if root_dir is None:
            root_dir = os.path.join(os.sep, "data", "visitor")
        session_dir = path_utils.markdir(
            os.path.join(root_dir, proposal, beamline, session)
        )
        if beamline not in _BEAMLINE_NAME_TO_DIR:
            return session_dir
        # Sometimes the directory name is equal to the beamline name and sometimes not
        if os.path.exists(session_dir):
            return session_dir
        bldirname = _BEAMLINE_NAME_TO_DIR[beamline]
        return path_utils.markdir(os.path.join(root_dir, proposal, bldirname, session))
    _raise_raw_data_format_error(raw_data_format)


def parse_session_dir(
    session_dir: str, raw_data_format: str = "esrfv3"
) -> Tuple[Optional[str], Optional[str], Optional[str]]:
    """Get proposal, beamline and session name from the session directory."""
    proposal, beamline_dir, session = path_utils.split(session_dir)[-3:]
    if not session.isdigit():
        return None, None, None
    beamline = _BEAMLINE_DIR_TO_NAME.get(beamline_dir, beamline_dir)
    return proposal, beamline, session


def get_raw_data_dir(session_dir: str, raw_data_format: str = "esrfv3") -> str:
    """Get the raw data directory from proposal, beamline and session name.
    This is the directory when Bliss saves the raw data.
    """
    if raw_data_format in ("esrfv3", "id16bspec", "mx"):
        return path_utils.markdir(os.path.join(session_dir, "RAW_DATA"))
    if raw_data_format == "esrfv2":
        return path_utils.markdir(os.path.join(session_dir, "raw"))
    if raw_data_format == "esrfv1":
        return path_utils.markdir(session_dir)
    _raise_raw_data_format_error(raw_data_format)


def get_dataset_filters(
    raw_root_dir: str, raw_data_format: str = "esrfv3"
) -> List[str]:
    """Get the dataset directory search filters from the raw data directory."""
    if raw_data_format in ("esrfv1", "esrfv2", "esrfv3", "id16bspec"):
        return [path_utils.markdir(os.path.join(raw_root_dir, "*", "*"))]
    elif raw_data_format == "mx":
        filters = []
        for root, dirs, files in os.walk(raw_root_dir):
            if MX_METADATA_FILENAME in files:
                filters.append(path_utils.markdir(root))
        return filters
    _raise_raw_data_format_error(raw_data_format)


def get_raw_dataset_name(
    dataset_dir: str, raw_data_format: str = "esrfv3"
) -> Optional[str]:
    """Get the raw data dataset name from the dataset directory."""
    if raw_data_format in ("esrfv1", "esrfv2", "esrfv3"):
        collection, collection_dataset = path_utils.split(dataset_dir)[-2:]
        if not collection_dataset.startswith(collection):
            return None
        return collection_dataset[len(collection) + 1 :]
    if raw_data_format in ("id16bspec", "mx"):
        return path_utils.split(dataset_dir)[-1]
    _raise_raw_data_format_error(raw_data_format)


def get_raw_dataset_metadata(
    dataset_dir: str, raw_data_format: str = "esrfv3"
) -> Dict[str, str]:
    """Get dataset info from the raw dataset directory."""
    if raw_data_format in ("esrfv1", "esrfv2", "esrfv3"):
        return _raw_dataset_metadata_esrf(dataset_dir)
    if raw_data_format == "id16bspec":
        return _raw_dataset_metadata_id16bspec(dataset_dir)
    if raw_data_format == "mx":
        return _raw_dataset_metadata_mx(dataset_dir)
    _raise_raw_data_format_error(raw_data_format)


def _raise_raw_data_format_error(raw_data_format: str) -> None:
    if raw_data_format in _DATA_FORMATS:
        raise RuntimeError(f"Implementation error for '{raw_data_format}'")
    else:
        raise NotImplementedError(
            f"Raw data format '{raw_data_format}' is not supported"
        )


def _raw_dataset_metadata_esrf(dataset_dir: str) -> Dict[str, str]:
    basename = path_utils.basename(dataset_dir)
    dataset_file = os.path.join(dataset_dir, f"{basename}.h5")
    if not os.path.exists(dataset_file):
        raise FileNotFoundError("HDF5 file does not exist")

    dataset_metadata = dict()
    enddate = None
    try:
        with h5py.File(dataset_file, "r", locking=False) as f:
            if not _is_bliss_raw_dataset_file(f):
                raise ValueError("HDF5 file not created by Bliss")
            startdate = f.attrs.get("file_time")
            for scan in map(str, sorted(map(float, list(f)))):
                sample_name = _read_hdf5_dataset(
                    f, f"/{scan}/sample/name", default=None
                )
                if sample_name is not None:
                    dataset_metadata["Sample_name"] = sample_name
                enddate = _read_hdf5_dataset(f, f"/{scan}/end_time", default=enddate)
    except Exception as e:
        raise RuntimeError(f"HDF5 reading error ({e})") from e

    if startdate is not None:
        dataset_metadata["startDate"] = startdate
    if enddate is not None:
        dataset_metadata["endDate"] = enddate

    return dataset_metadata


def _raw_dataset_metadata_id16bspec(dataset_dir: str) -> Dict[str, str]:
    dataset_metadata = dict()

    proposal, _, _, _, sample_name, dataset = path_utils.split(dataset_dir)[-6:]
    filename = f"{proposal}-{sample_name}-{dataset}.h5"
    dataset_file = os.path.join(dataset_dir, filename)

    if not os.path.exists(dataset_file):
        raise FileNotFoundError("HDF5 file does not exist")

    startdate = None
    enddate = None
    try:
        with h5py.File(dataset_file, "r", locking=False) as f:
            for name in f:
                entry = f[name]
                try:
                    startdate = _read_hdf5_dataset(entry, "start_time", default=None)
                    enddate = _read_hdf5_dataset(entry, "end_time", default=None)
                except KeyError as e:
                    raise ValueError(f"Time could not be read from HDF5 ({e})") from e
                break
    except Exception as e:
        raise RuntimeError(f"HDF5 reading error ({e})") from e

    if startdate is not None:
        dataset_metadata["startDate"] = startdate
    if enddate is not None:
        dataset_metadata["endDate"] = enddate
    dataset_metadata["Sample_name"] = sample_name
    return dataset_metadata


def _raw_dataset_metadata_mx(dataset_dir: str) -> Dict[str, str]:
    """Read metadata from MX_METADATA_FILENAME for 'mx' format."""
    metadata_file = os.path.join(dataset_dir, MX_METADATA_FILENAME)

    if not os.path.isfile(metadata_file):
        raise FileNotFoundError(f"{MX_METADATA_FILENAME} not found in {dataset_dir}")

    with open(metadata_file, "r") as f:
        metadata = json.load(f)

    return {key: str(value) for key, value in metadata.items() if value is not None}


def _is_bliss_raw_dataset_file(f: h5py.File) -> bool:
    return f.attrs.get("creator", "").lower() in ("bliss", "blissdata", "blisswriter")


def _read_hdf5_dataset(parent: h5py.Group, name: str, default=None) -> Any:
    try:
        value = parent[name][()]
    except KeyError:
        return default
    try:
        return value.decode()
    except AttributeError:
        pass
    return value



import pytest
import json


def add_files(client, value):
    client.add_files(
        dataset_id=value["datasetId"],
    )


def test_add_files(icat_add_files_client):
    client, messages = icat_add_files_client
    client.check_health()
    expected = {
        "datasetId": 1234,
    }
    add_files(client, expected)
    message = messages.get(timeout=10)
    message = json.loads(message)
    assert message == expected
    assert messages.empty()


def test_add_files_raises_error_if_dataset_id_is_missing(
    icat_add_files_client,
):
    client, messages = icat_add_files_client
    with pytest.raises(AssertionError, match="ICAT requires the datasetId"):
        client.add_files(
            dataset_id=None,
        )



import os
from pathlib import Path

import pytest

from ..concurrency import Empty
from .utils.message import assert_dataset_message
from .utils.message import assert_investigation_message
from .utils import generate


def test_start_investigation(icat_metadata_client):
    client, messages = icat_metadata_client
    client.check_health()
    client.start_investigation(proposal="hg123", beamline="id00")
    message = messages.get(timeout=10)
    assert messages.empty()

    expected = {"investigation": {"experiment": "hg123", "instrument": "id00"}}
    assert_investigation_message(message, expected)


def test_start_bad_investigation(icat_metadata_client):
    client, messages = icat_metadata_client
    client.check_health()
    client.start_investigation(proposal="hg666", beamline="id00")
    with pytest.raises(Empty):
        messages.get(timeout=2)


def test_send_metadata(icat_metadata_client):
    client, messages = icat_metadata_client
    client.send_metadata(
        proposal="hg123",
        beamline="id00",
        dataset="datasetname",
        path=_dummy_path("dataset"),
        metadata={"Sample_name": "samplename", "field1": "value1", "field2": [1, 2, 3]},
    )
    message = messages.get(timeout=10)
    assert messages.empty()

    expected = {
        "dataset": {
            "@complete": "true",
            "@xmlns": {"tns": "http://www.esrf.fr/icat"},
            "instrument": "id00",
            "investigation": "hg123",
            "location": str(_dummy_path("dataset")),
            "name": "datasetname",
            "parameter": [
                {"name": "Sample_name", "value": "samplename"},
                {"name": "field1", "value": "value1"},
                {"name": "field2", "value": "1,2,3"},
            ],
            "sample": {"name": "samplename"},
        }
    }
    assert_dataset_message(message, expected)


def test_send_metadata_via_file(icat_metadata_client, tmpdir):
    store_filename = tmpdir / "test.xml"

    client, messages = icat_metadata_client
    client.store_metadata(
        str(store_filename),
        proposal="hg123",
        beamline="id00",
        dataset="datasetname",
        path=_dummy_path("dataset"),
        metadata={"Sample_name": "samplename", "field1": "value1", "field2": [1, 2, 3]},
    )

    with pytest.raises(Empty):
        message = messages.get(timeout=1)

    assert store_filename.exists()

    client.send_metadata_from_file(str(store_filename))

    message = messages.get(timeout=10)
    assert messages.empty()

    expected = {
        "dataset": {
            "@complete": "true",
            "@xmlns": {"tns": "http://www.esrf.fr/icat"},
            "instrument": "id00",
            "investigation": "hg123",
            "location": str(_dummy_path("dataset")),
            "name": "datasetname",
            "parameter": [
                {"name": "Sample_name", "value": "samplename"},
                {"name": "field1", "value": "value1"},
                {"name": "field2", "value": "1,2,3"},
            ],
            "sample": {"name": "samplename"},
        }
    }
    assert_dataset_message(message, expected)


def test_send_missing_data(icat_metadata_client):
    client, messages = icat_metadata_client
    with pytest.raises(AssertionError, match="ICAT requires the beamline name"):
        client.send_metadata(
            proposal=None,
            beamline=None,
            dataset=None,
            path=None,
            metadata=None,
        )


def test_send_missing_metadata(icat_metadata_client):
    client, messages = icat_metadata_client
    with pytest.raises(
        AssertionError, match="ICAT metadata field 'Sample_name' is missing"
    ):
        client.send_metadata(
            proposal="hg123",
            beamline="id00",
            dataset="datasetname",
            path=_dummy_path("dataset"),
            metadata={},
        )


def test_send_metadata_with_machine_software(icat_metadata_client):
    client, messages = icat_metadata_client
    client.send_metadata(
        proposal="hg123",
        beamline="id00",
        dataset="datasetname",
        path=_dummy_path("dataset"),
        metadata={
            "Sample_name": "samplename",
            "field1": "value1",
            "field2": [1, 2, 3],
            "machine": "mymachine",
            "software": "mysoftware_version",
        },
    )
    message = messages.get(timeout=10)
    assert messages.empty()

    expected = {
        "dataset": {
            "@complete": "true",
            "@xmlns": {"tns": "http://www.esrf.fr/icat"},
            "instrument": "id00",
            "investigation": "hg123",
            "location": str(_dummy_path("dataset")),
            "name": "datasetname",
            "parameter": [
                {"name": "Sample_name", "value": "samplename"},
                {"name": "field1", "value": "value1"},
                {"name": "field2", "value": "1,2,3"},
                {"name": "machine", "value": "mymachine"},
                {"name": "software", "value": "mysoftware_version"},
            ],
            "sample": {"name": "samplename"},
        }
    }
    assert_dataset_message(message, expected)


def test_reschedule_investigation(icat_metadata_client):
    client, messages = icat_metadata_client
    client.check_health()
    investigation_id = generate.investigation_id()
    client.reschedule_investigation(investigation_id=investigation_id)
    message = messages.get(timeout=10)
    assert messages.empty()

    expected = {
        "investigation": {
            "experiment": None,
            "instrument": None,
            "investigationId": investigation_id,
        }
    }
    assert_investigation_message(message, expected)


@pytest.mark.parametrize("path_defect", ["none", "trailing_slash", "up"])
def test_send_metadata_path_defects(icat_metadata_client, path_defect):
    expected_path = _dummy_path("dataset")
    expected_input_datasets = [_dummy_path("dataset1"), _dummy_path("dataset2")]

    if path_defect == "trailing_slash":
        path = _add_trailing_slash(expected_path)
        assert path.endswith(os.path.sep)
        input_datasets = [_add_trailing_slash(s) for s in expected_input_datasets]
    elif path_defect == "up":
        path = _add_up(expected_path)
        input_datasets = [_add_up(s) for s in expected_input_datasets]
    else:
        path = expected_path
        input_datasets = expected_input_datasets

    client, messages = icat_metadata_client
    client.send_metadata(
        proposal="hg123",
        beamline="id00",
        dataset="datasetname",
        path=path,
        metadata={
            "Sample_name": "samplename",
            "field1": "value1",
            "field2": [1, 2, 3],
            "input_datasets": input_datasets,
        },
    )

    message = messages.get(timeout=10)
    assert messages.empty()

    expected_path = str(expected_path)
    input_datasets = ",".join([str(s) for s in expected_input_datasets])

    expected = {
        "dataset": {
            "@complete": "true",
            "@xmlns": {"tns": "http://www.esrf.fr/icat"},
            "instrument": "id00",
            "investigation": "hg123",
            "location": str(expected_path),
            "name": "datasetname",
            "parameter": [
                {"name": "Sample_name", "value": "samplename"},
                {"name": "field1", "value": "value1"},
                {"name": "field2", "value": "1,2,3"},
                {"name": "input_datasets", "value": input_datasets},
            ],
            "sample": {"name": "samplename"},
        }
    }
    assert_dataset_message(message, expected)


def _add_trailing_slash(path: Path) -> str:
    return str(path / "_")[:-1]


def _add_up(path: Path) -> Path:
    return path / "_" / ".."


def _dummy_path(dirname: str) -> Path:
    return Path.home() / dirname



import socket
from typing import Optional
from importlib.metadata import version as get_version

release = get_version("pyicat_plus")


def check_parameter_exists(dataset, param_name):
    return any(param["name"] == param_name for param in dataset["parameter"])


def assert_dataset_message(message: dict, expected: dict):
    expected["dataset"]["startDate"] = message["dataset"]["startDate"]
    expected["dataset"]["endDate"] = message["dataset"]["endDate"]
    expected["dataset"]["parameter"].append(_get_parameter(message, "startDate"))
    expected["dataset"]["parameter"].append(_get_parameter(message, "endDate"))
    if not check_parameter_exists(expected["dataset"], "machine"):
        expected["dataset"]["parameter"].append(
            {"name": "machine", "value": socket.getfqdn()}
        )
    if not check_parameter_exists(expected["dataset"], "software"):
        expected["dataset"]["parameter"].append(
            {"name": "software", "value": "pyicat-plus_v" + release}
        )
    expected["dataset"]["parameter"] = sorted(
        expected["dataset"]["parameter"], key=lambda adict: adict["name"]
    )
    message["dataset"]["parameter"] = sorted(
        message["dataset"]["parameter"], key=lambda adict: adict["name"]
    )
    assert message == expected


def assert_investigation_message(message: dict, expected: dict):
    expected["investigation"]["startDate"] = message["investigation"]["startDate"]
    assert message == expected


def _get_parameter(root: dict, parameter_name: str) -> Optional[dict]:
    for parameter in root["dataset"]["parameter"]:
        if parameter["name"] == parameter_name:
            return parameter



import os
import logging
import datetime
import dataclasses
from typing import Optional, Dict, Any, List, Set, Generator

from . import path_utils
from ..client.interface import Dataset as IcatDataset


logger = logging.getLogger("ICAT SYNC")


@dataclasses.dataclass(eq=True)
class Dataset:
    path: str  # Path of the directory
    proposal: str
    beamline: str
    name: Optional[str]
    raw_root_dir: str
    status_reason: List[str]
    startdate: Optional[datetime.datetime] = None  # From HDF5
    enddate: Optional[datetime.datetime] = None  # From HDF5
    metadata: Optional[dict] = (
        None  # From HDF5 (times are clipped to the session time slot)
    )
    icat_dataset: Optional[IcatDataset] = None  # From ICAT

    def __post_init__(self) -> None:
        self.path = path_utils.markdir(self.path)

    def as_dict(self) -> Dict[str, Any]:
        return dataclasses.asdict(self)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "Dataset":
        """Factory method to create a Dataset instance from a dictionary."""
        data = data.copy()
        icat_dataset = data.get("icat_dataset")
        if icat_dataset is not None:
            data["icat_dataset"] = IcatDataset.from_dict(icat_dataset)
        return cls(**data)

    def is_invalid(self) -> bool:
        return self.name is None or not os.path.isdir(self.path)

    def has_metadata(self) -> bool:
        return self.metadata is not None

    def to_be_uploaded(self) -> bool:
        return not self.is_registered() and os.path.exists(self.metadata_file)

    def is_registered(self) -> bool:
        return self.icat_dataset is not None

    def is_non_empty_dir(self) -> bool:
        if not os.path.isdir(self.path):
            return False
        return bool(os.listdir(self.path))

    def is_registered_without_files(self) -> bool:
        return (
            self.is_registered()
            and self.icat_dataset.dataset_metadata.file_count == 0
            and self.is_non_empty_dir()
        )

    def get_status(self) -> str:
        # Decreasing order of registration:
        if self.is_registered_without_files():
            return "registered_without_files"
        if self.is_registered():
            return "registered"
        if self.to_be_uploaded():
            return "not_uploaded"
        # Not registered of invalid
        if self.is_invalid() or not self.has_metadata():
            return "invalid"
        return "unregistered"

    @property
    def metadata_file(self) -> str:
        """File stored by Bliss when ending the proposal without confirmation of registration"""
        return os.path.join(
            self.raw_root_dir, "__icat__", f"{path_utils.basename(self.path)}.xml"
        )


@dataclasses.dataclass(eq=True)
class IcatInvestigation:
    id: str
    url: str
    search_url: str
    registered_datasets: List[IcatDataset]
    startdate: Optional[datetime.datetime] = None
    enddate: Optional[datetime.datetime] = None

    def as_dict(self) -> Dict[str, Any]:
        return dataclasses.asdict(self)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "IcatInvestigation":
        """Factory method to create an IcatInvestigation instance from a dictionary."""
        data = data.copy()
        data["registered_datasets"] = [
            IcatDataset.from_dict(dataset_data)
            for dataset_data in data["registered_datasets"]
        ]
        return cls(**data)

    @property
    def has_timeslot(self) -> bool:
        return self.startdate is not None

    @property
    def open_ended(self) -> bool:
        return self.enddate is None

    @property
    def started(self) -> bool:
        if not self.has_timeslot:
            return False
        now = datetime.datetime.now().astimezone()
        startdate = self.startdate - datetime.timedelta(days=2)
        return now >= startdate

    def _start_from_now(self) -> str:
        if not self.has_timeslot:
            return ""
        now = datetime.datetime.now().astimezone()
        if now >= self.startdate:
            return f"(Started {now - self.startdate} ago)"
        else:
            return f"(Will start in {self.startdate - now})"

    @property
    def ended(self) -> bool:
        if self.open_ended:
            return False
        now = datetime.datetime.now().astimezone()
        enddate = self.enddate + datetime.timedelta(days=2)
        return now >= enddate

    def _end_from_now(self) -> str:
        if self.open_ended:
            return ""
        now = datetime.datetime.now().astimezone()
        if now >= self.enddate:
            return f"(Ended {now - self.enddate} ago)"
        else:
            return f"(Will end in {self.enddate - now})"

    @property
    def ongoing(self) -> bool:
        return self.started and not self.ended

    @property
    def duration(self) -> Optional[datetime.timedelta]:
        if not self.open_ended:
            return self.enddate - self.startdate

    @property
    def timeslot(self) -> str:
        if self.open_ended:
            return f"{self.startdate} - ???"
        return f"{self.startdate} - {self.enddate} ({self.duration})"

    def pprint(self) -> None:
        logger.debug("Investigation ID: %s", self.id)
        logger.debug(" URL: %s", self.url)
        logger.debug(" Search URL: %s", self.search_url)

        if not self.has_timeslot:
            logger.debug("Invalid investigation (no start date)")
            return
        logger.debug(" Start time: %s %s", self.startdate, self._start_from_now())

        if self.open_ended:
            logger.debug(" Duration:", "open-ended (not official)")
            return
        logger.debug(" End time: %s %s", self.enddate, self._end_from_now())

        if self.ongoing:
            logger.debug(" Duration: %s (WARNING: ONGOING)", self.duration)
        else:
            logger.debug(" Duration: %s", self.duration)


@dataclasses.dataclass(eq=True)
class ExperimentalSession:
    session_dir: str
    raw_root_dir: str  # RAW_DATA directory
    raw_data_format: str  # e.g. "esrfv3"
    proposal: str
    beamline: str
    session: str
    startdate: datetime.date  # from the directory name
    search_url: str
    datasets: Dict[str, List[Dataset]]  # from browsing the root directory
    icat_investigation: Optional[IcatInvestigation] = None

    DATASET_STATUSES = (
        "unregistered",
        "not_uploaded",
        "registered",
        "invalid",
        "registered_without_files",
    )

    def __post_init__(self) -> None:
        self.datasets = {
            status: self.datasets.get(status, list())
            for status in self.DATASET_STATUSES
        }
        self.session_dir = path_utils.markdir(self.session_dir)
        self.raw_root_dir = path_utils.markdir(self.raw_root_dir)

    def as_dict(self) -> Dict[str, Any]:
        return dataclasses.asdict(self)

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "ExperimentalSession":
        data = data.copy()

        data["datasets"] = {
            key: [Dataset.from_dict(dataset_data) for dataset_data in dataset_list]
            for key, dataset_list in data["datasets"].items()
        }

        icat_investigation = data.get("icat_investigation")
        if icat_investigation is not None:
            data["icat_investigation"] = IcatInvestigation.from_dict(icat_investigation)

        return cls(**data)

    @classmethod
    def allow_unsupervised_upload(cls, dataset_status: str) -> bool:
        if dataset_status not in cls.DATASET_STATUSES:
            raise ValueError(
                f"'{dataset_status}' is not as valid dataset status. Valid statuses are {list(cls.DATASET_STATUSES)}"
            )
        return dataset_status == "registered_without_files"

    @property
    def in_icat_investigation(self) -> Optional[bool]:
        if self.icat_investigation is None:
            return
        if not self.icat_investigation.has_timeslot:
            return False
        inside_timeslot = self.startdate >= self.icat_investigation.startdate.date()
        if not self.icat_investigation.open_ended:
            inside_timeslot &= self.startdate <= self.icat_investigation.enddate.date()
        return inside_timeslot

    def add_dataset(self, dataset: Dataset) -> None:
        self.datasets[dataset.get_status()].append(dataset)

    def iter_datasets(self) -> Generator[Dataset, None, None]:
        for lst in self.datasets.values():
            yield from lst

    def dataset_statuses(self) -> Set[str]:
        return set(status for status, lst in self.datasets.items() if lst)

    def dataset_status_counts(self) -> Dict[str, int]:
        return {status: len(lst) for status, lst in self.datasets.items()}

    def tabular_data(self) -> Dict[str, str]:
        data = self._init_tabular_data()
        self._add_icat_tabular_data(data)
        return data

    def _init_tabular_data(self) -> Dict[str, str]:
        status_counts = {
            status: str(counts)
            for status, counts in self.dataset_status_counts().items()
        }
        return {
            "beamline": self.beamline,
            "proposal": self.proposal,
            "session": str(self.startdate),
            "url": "",
            "search_url": self.search_url,
            "path": self.raw_root_dir,
            **status_counts,
            "timeslot": "",
        }

    def _add_icat_tabular_data(self, data: Dict[str, str]) -> None:
        if self.icat_investigation is None:
            return
        data["url"] = self.icat_investigation.url
        data["search_url"] = self.icat_investigation.search_url

        if self.in_icat_investigation:
            data["session"] += " (NOT INSIDE TIMESLOT)"
        if not self.icat_investigation.started:
            data["timeslot"] = f"FUTURE: {self.icat_investigation.timeslot}"
        elif self.icat_investigation.ended:
            data["timeslot"] = f"FINISHED: {self.icat_investigation.timeslot}"
        elif self.icat_investigation.open_ended:
            data["timeslot"] = f"INDEFINITE: {self.icat_investigation.timeslot}"
        elif self.icat_investigation.ongoing:
            data["timeslot"] = f"ONGOING: {self.icat_investigation.timeslot}"
        else:
            raise RuntimeError("Unknown session time state")

        return data

    def pprint(self) -> None:
        separator = "--------------------------------"
        logger.debug("")
        logger.debug(separator)
        self._pprint_datasets()
        logger.debug("")
        self._pprint_raw_session()
        logger.debug("")
        self._pprint_icat_investigation()
        logger.debug("")
        logger.debug(separator)
        logger.debug("")

    def _pprint_raw_session(self) -> None:
        logger.debug("Directory: %s", self.raw_root_dir)
        if self.in_icat_investigation:
            logger.debug(" Start time: %s", self.startdate)
        else:
            logger.debug(
                " Start time: %s (WARNING: not inside time slot!)", self.startdate
            )

    def _pprint_icat_investigation(self) -> None:
        if self.icat_investigation is None:
            logger.debug("No corresponding investigation")
            logger.debug(" Search URL: %s", self.search_url)
        else:
            self.icat_investigation.pprint()

    def _pprint_datasets(self) -> None:
        _print_dataset_summary("Registered datasets", self.datasets["registered"])
        _print_dataset_summary("Invalid datasets", self.datasets["invalid"])
        _print_dataset_summary("Unregistered datasets", self.datasets["unregistered"])
        _print_dataset_summary("Datasets not uploaded", self.datasets["not_uploaded"])
        _print_dataset_summary(
            "Registered datasets without files",
            self.datasets["registered_without_files"],
        )

        logger.debug("")
        if (
            self.datasets["invalid"]
            or self.datasets["not_uploaded"]
            or self.datasets["unregistered"]
            or self.datasets["registered_without_files"]
        ):
            logger.debug("Datasets (TODO):")
        elif not self.datasets["registered"]:
            logger.debug("Datasets (EMPTY):")
        else:
            logger.debug("Datasets (OK):")
        logger.debug(" %s registered", len(self.datasets["registered"]))
        logger.debug(" %s invalid", len(self.datasets["invalid"]))
        logger.debug(" %s not uploaded", len(self.datasets["not_uploaded"]))
        logger.debug(" %s unregistered", len(self.datasets["unregistered"]))
        logger.debug(
            " %s registered without files",
            len(self.datasets["registered_without_files"]),
        )


def _print_dataset_summary(title: str, datasets: List[Dataset]):
    if not datasets:
        return
    logger.debug("")
    logger.debug("%s:", title)
    for dataset in datasets:
        logger.debug("  %s", dataset.path)
        for s in dataset.status_reason:
            logger.debug("   %s", s)



import os
import socket
import datetime
from typing import Union, Optional

import xml.etree.ElementTree as etree

from .serialize import serialize_metadata
from importlib.metadata import version as get_version

release = get_version("pyicat_plus")

ICAT_NAMESPACE_URL = "http://www.esrf.fr/icat"

etree.register_namespace("tns", ICAT_NAMESPACE_URL)


def root_node(name: str, **kw):
    return etree.Element(f"{{{ICAT_NAMESPACE_URL}}}{name}", **kw)


def child_node(parent, name: str, **kw):
    return etree.SubElement(parent, f"{{{ICAT_NAMESPACE_URL}}}{name}", **kw)


def encode_node_data(data) -> str:
    text = serialize_metadata(data)
    if not isinstance(text, str):
        raise ValueError(data)
    return text


def data_node(parent, name: str, data, **kw):
    node = child_node(parent, name, **kw)
    node.text = encode_node_data(data)


def parameter_node(parent, name: str, value, **kw):
    node = child_node(parent, "parameter", **kw)
    data_node(node, "name", name)
    data_node(node, "value", value)


def dataset_as_xml(
    beamline: str, proposal: str, dataset: str, path: str, metadata: dict = None
):
    assert beamline, "ICAT requires the beamline name"
    assert proposal, "ICAT requires the proposal name"
    assert dataset, "ICAT requires the dataset name"
    assert path, "ICAT requires the dataset path"
    if metadata is None:
        metadata = dict()

    # Required metadata
    assert "Sample_name" in metadata, "ICAT metadata field 'Sample_name' is missing"

    # Metadata with defaults
    if "startDate" not in metadata:
        metadata["startDate"] = datetime.datetime.now().astimezone()
    if "endDate" not in metadata:
        metadata["endDate"] = datetime.datetime.now().astimezone()

    metadata.setdefault("machine", socket.getfqdn())
    metadata.setdefault("software", "pyicat-plus_v" + release)

    # Normalize fields
    path = os.path.normpath(path)
    input_datasets = metadata.get("input_datasets", None)
    if input_datasets:
        metadata["input_datasets"] = [
            os.path.normpath(input_dataset) for input_dataset in input_datasets
        ]

    # Create XML tree
    root = root_node("dataset", attrib={"complete": "true"})
    data_node(root, "investigation", proposal)
    data_node(root, "instrument", beamline)
    data_node(root, "name", dataset)
    data_node(root, "location", path)

    metadata = serialize_metadata(metadata)
    for name, value in metadata.items():
        parameter_node(root, name, value)
        # Metadata included in the XML tree
        if name == "Sample_name":
            sample = child_node(root, "sample")
            data_node(sample, "name", value)
        elif name == "startDate":
            data_node(root, "startDate", value)
        elif name == "endDate":
            data_node(root, "endDate", value)

    return root


def investigation_as_xml(
    beamline: str,
    proposal: str,
    start_datetime: Optional[Union[datetime.datetime, str]] = None,
    end_datetime: Optional[Union[datetime.datetime, str]] = None,
    investigation_id: Optional[str] = None,
):
    root = root_node("investigation")
    data_node(root, "experiment", proposal)
    data_node(root, "instrument", beamline)
    if investigation_id is not None:
        data_node(root, "investigationId", investigation_id)
    if start_datetime is None:
        start_datetime = datetime.datetime.now().astimezone()
    data_node(root, "startDate", start_datetime)
    if end_datetime is not None:
        data_node(root, "endDate", end_datetime)
    return root



import os
import re
import time
import logging
import datetime
from typing import Generator, Tuple, Optional, Callable
from contextlib import contextmanager

import h5py

from ..apps import sync_raw
from ..utils import sync_types
from ..utils import path_utils
from ..client.main import IcatClient


def test_unregistered_datasets(tmpdir, icat_main_client):
    client, messages = icat_main_client
    root_dir = str(tmpdir)

    exp_session_test = _init_unregistered_experiment(
        client, messages, root_dir, "hg333", "id99"
    )

    args = client, "hg333", "id99", "20231019"
    kwargs = {"root_dir": root_dir, "raw_data_format": "esrfv3"}

    exp_session = sync_raw.sync_session(*args, **kwargs)
    assert len(exp_session.datasets["registered"]) == 0
    assert len(exp_session.datasets["unregistered"]) == 6

    if exp_session != exp_session_test:
        assert exp_session.as_dict() == exp_session_test.as_dict()

    first_datasets = exp_session_test.datasets["unregistered"][:3]
    last_datasets = exp_session_test.datasets["unregistered"][3:]

    for dataset in first_datasets:
        _store_dataset(client, dataset)
        messages.get(timeout=10)

    exp_session = sync_raw.sync_session(*args, **kwargs)
    assert len(exp_session.datasets["registered"]) == 3
    assert len(exp_session.datasets["unregistered"]) == 3

    for dataset in last_datasets:
        _store_dataset(client, dataset)
        messages.get(timeout=10)

    exp_session = sync_raw.sync_session(*args, **kwargs)
    assert len(exp_session.datasets["registered"]) == 6
    assert len(exp_session.datasets["unregistered"]) == 0


def test_unregistered_datasets_content(tmpdir, icat_main_client):
    client, messages = icat_main_client

    root_dir = str(tmpdir)
    _ = _init_unregistered_experiment(client, messages, root_dir, "hg333", "id99")

    exp_session = sync_raw.sync_session(
        client, "hg333", "id99", "20231019", root_dir=root_dir, raw_data_format="esrfv3"
    )

    unregistered = [dset.as_dict() for dset in exp_session.datasets["unregistered"]]

    date_to_string = _astimezone_string(datetime.timezone.utc, iso=True)
    for dataset in unregistered:
        dataset["path"] = os.path.relpath(dataset["path"], root_dir)
        dataset["raw_root_dir"] = os.path.relpath(dataset["raw_root_dir"], root_dir)

        dataset["startdate"] = date_to_string(dataset["startdate"])
        dataset["enddate"] = date_to_string(dataset["enddate"])

        metadata = dataset["metadata"]
        metadata["startDate"] = date_to_string(metadata["startDate"])
        metadata["endDate"] = date_to_string(metadata["endDate"])

    expected = [
        {
            "beamline": "id99",
            "enddate": "2023-10-19T21:00:00+00:00",
            "icat_dataset": None,
            "metadata": {
                "Sample_name": "too_soon_sample",
                "endDate": "2023-10-19T21:00:00+00:00",
                "startDate": "2023-10-19T06:00:01+00:00",
            },
            "name": "0001",
            "path": "hg333/id99/20231019/RAW_DATA/too_soon_collection/too_soon_collection_0001",
            "proposal": "hg333",
            "raw_root_dir": "hg333/id99/20231019/RAW_DATA",
            "startdate": "2023-10-19T05:00:00+00:00",
            "status_reason": [
                "started 1:00:00 before the start of the session",
            ],
        },
        {
            "beamline": "id99",
            "enddate": "2023-10-20T16:00:00+00:00",
            "icat_dataset": None,
            "metadata": {
                "Sample_name": "first_sample",
                "endDate": "2023-10-20T16:00:00+00:00",
                "startDate": "2023-10-20T00:00:00+00:00",
            },
            "name": "0001",
            "path": "hg333/id99/20231019/RAW_DATA/first_collection/first_collection_0001",
            "proposal": "hg333",
            "raw_root_dir": "hg333/id99/20231019/RAW_DATA",
            "startdate": "2023-10-20T00:00:00+00:00",
            "status_reason": [],
        },
        {
            "beamline": "id99",
            "enddate": "2023-10-21T09:00:00+00:00",
            "icat_dataset": None,
            "metadata": {
                "Sample_name": "first_sample",
                "endDate": "2023-10-21T09:00:00+00:00",
                "startDate": "2023-10-20T17:00:00+00:00",
            },
            "name": "0002",
            "path": "hg333/id99/20231019/RAW_DATA/first_collection/first_collection_0002",
            "proposal": "hg333",
            "raw_root_dir": "hg333/id99/20231019/RAW_DATA",
            "startdate": "2023-10-20T17:00:00+00:00",
            "status_reason": [],
        },
        {
            "beamline": "id99",
            "enddate": "2023-10-22T19:00:00+00:00",
            "icat_dataset": None,
            "metadata": {
                "Sample_name": "second_sample",
                "endDate": "2023-10-22T19:00:00+00:00",
                "startDate": "2023-10-22T03:00:00+00:00",
            },
            "name": "0001",
            "path": "hg333/id99/20231019/RAW_DATA/second_collection/second_collection_0001",
            "proposal": "hg333",
            "raw_root_dir": "hg333/id99/20231019/RAW_DATA",
            "startdate": "2023-10-22T03:00:00+00:00",
            "status_reason": [],
        },
        {
            "beamline": "id99",
            "enddate": "2023-10-23T12:00:00+00:00",
            "icat_dataset": None,
            "metadata": {
                "Sample_name": "second_sample",
                "endDate": "2023-10-23T12:00:00+00:00",
                "startDate": "2023-10-22T20:00:00+00:00",
            },
            "name": "0002",
            "path": "hg333/id99/20231019/RAW_DATA/second_collection/second_collection_0002",
            "proposal": "hg333",
            "raw_root_dir": "hg333/id99/20231019/RAW_DATA",
            "startdate": "2023-10-22T20:00:00+00:00",
            "status_reason": [],
        },
        {
            "beamline": "id99",
            "enddate": "2023-10-24T07:00:00+00:00",
            "icat_dataset": None,
            "metadata": {
                "Sample_name": "too_late_collection",
                "endDate": "2023-10-24T05:59:59+00:00",
                "startDate": "2023-10-23T13:00:00+00:00",
            },
            "name": "0001",
            "path": "hg333/id99/20231019/RAW_DATA/too_late_collection/too_late_collection_0001",
            "proposal": "hg333",
            "raw_root_dir": "hg333/id99/20231019/RAW_DATA",
            "startdate": "2023-10-23T13:00:00+00:00",
            "status_reason": [
                "ended 1:00:00 after the end of the session",
            ],
        },
    ]

    for dataset in expected:
        for k in ("path", "raw_root_dir"):
            dataset[k] = dataset[k].replace("/", os.path.sep)

    assert unregistered == expected


def test_missing_files_datasets(tmpdir, icat_main_client, icat_add_files_client):
    client, messages = icat_main_client
    client_add, messages_add = icat_add_files_client
    root_dir = str(tmpdir)

    _ = _init_empty_files_experiment(client, messages, root_dir, "md444", "id99")

    args = client, "md444", "id99", "20230201"
    kwargs = {"root_dir": root_dir, "raw_data_format": "esrfv3"}

    exp_session = sync_raw.sync_session(*args, **kwargs)
    assert len(exp_session.datasets["registered"]) == 1
    assert len(exp_session.datasets["registered_without_files"]) == 1
    assert len(exp_session.datasets["unregistered"]) == 0

    for dataset in exp_session.datasets["registered_without_files"]:
        client_add.add_files(dataset.icat_dataset.icat_dataset_id)
        messages_add.get(timeout=10)

    exp_session = sync_raw.sync_session(*args, **kwargs)
    assert len(exp_session.datasets["registered"]) == 2
    assert len(exp_session.datasets["registered_without_files"]) == 0
    assert len(exp_session.datasets["unregistered"]) == 0


def test_session_serialization(tmpdir, icat_main_client):
    client, messages = icat_main_client

    root_dir = str(tmpdir)
    _ = _init_unregistered_experiment(client, messages, root_dir, "hg333", "id99")

    args = client, "hg333", "id99", "20231019"
    kwargs = {"root_dir": root_dir, "raw_data_format": "esrfv3"}

    exp_session = sync_raw.sync_session(*args, **kwargs)

    exp_session_copy = exp_session.from_dict(exp_session.as_dict())
    assert exp_session == exp_session_copy


def test_session_pprint(caplog, tmpdir, icat_main_client):
    client, messages = icat_main_client

    root_dir = str(tmpdir)
    _ = _init_unregistered_experiment(client, messages, root_dir, "hg333", "id99")

    args = client, "hg333", "id99", "20231019"
    kwargs = {"root_dir": root_dir, "raw_data_format": "esrfv3"}

    exp_session = sync_raw.sync_session(*args, **kwargs)

    for dataset in exp_session.datasets["unregistered"][:3]:
        _store_dataset(client, dataset)
        messages.get(timeout=10)

    exp_session = sync_raw.sync_session(*args, **kwargs)

    date_to_string = _astimezone_string(
        exp_session.icat_investigation.startdate.tzinfo, iso=False
    )
    startdate = date_to_string(exp_session.icat_investigation.startdate)
    enddate = date_to_string(exp_session.icat_investigation.enddate)

    expected = f"""
--------------------------------

Registered datasets:
  /hg333/id99/20231019/RAW_DATA/too_soon_collection/too_soon_collection_0001/
  /hg333/id99/20231019/RAW_DATA/first_collection/first_collection_0001/
  /hg333/id99/20231019/RAW_DATA/first_collection/first_collection_0002/

Invalid datasets:
  /hg333/id99/20231019/RAW_DATA/empty_file/empty_file_0001/
   Cannot extract dataset metadata: HDF5 reading error (HDF5 file not created by Bliss)

Unregistered datasets:
  /hg333/id99/20231019/RAW_DATA/second_collection/second_collection_0001/
  /hg333/id99/20231019/RAW_DATA/second_collection/second_collection_0002/
  /hg333/id99/20231019/RAW_DATA/too_late_collection/too_late_collection_0001/
   ended 1:00:00 after the end of the session

Datasets (TODO):
 3 registered
 1 invalid
 0 not uploaded
 3 unregistered
 0 registered without files

Directory: /hg333/id99/20231019/RAW_DATA/
 Start time: 2023-10-19

Investigation ID: 0
 URL: https://data.esrf.fr/investigation/0/datasets
 Search URL: https://data.esrf.fr/beamline/id99?search=hg333
 Start time: {startdate} <REPLACED>
 End time: {enddate} <REPLACED>
 Duration: 5 days, 0:00:00

--------------------------------
"""

    expected = re.sub(
        r"(/hg333[^\n]*)",
        lambda match: match.group(0).replace("/", os.path.sep),
        expected,
    )

    caplog.set_level(logging.DEBUG, logger="ICAT SYNC")
    exp_session.pprint()
    captured = "\n".join(caplog.messages)
    captured = captured.replace(root_dir, "")

    # Remove the sub-strings that depend on the time this test is run
    captured = re.sub(r"\(.*ago\)", "<REPLACED>", captured)

    assert captured == expected


class _ExperimentalSession(sync_types.ExperimentalSession):
    def create_dataset(
        self,
        collection: str,
        dataset: str,
        sample_name: str,
        startdate: datetime.datetime,
        duration: datetime.timedelta,
        status_reason: Tuple[str] = tuple(),
        has_content: bool = True,
    ) -> sync_types.Dataset:
        """Create Bliss dataset directory with HDF5 file. The file contains 3 scans or nothing."""
        with _init_dataset(self.raw_root_dir, collection, dataset) as nxroot:
            dataset_dir = os.path.dirname(nxroot.filename)
            if has_content:
                nxroot.attrs["creator"] = "Bliss"
                nxroot.attrs["file_time"] = startdate.isoformat()

                now, scan_duration, deadtime = _chunk_duration(
                    startdate, startdate + duration, 3
                )

                enddate = _save_scan(nxroot, now, scan_duration, sample_name)

                now += scan_duration + deadtime
                enddate = _save_scan(
                    nxroot, now, scan_duration, sample_name, failed=True
                )

                now += scan_duration + deadtime
                enddate = _save_scan(nxroot, now, scan_duration, sample_name)
            time.sleep(0.050)  # sync_raw sorts by creation time

        if has_content:
            medata_startdate = startdate
            medata_enddate = enddate

            # See sync_raw metadata extraction from HDF5
            if medata_enddate and medata_startdate <= self.icat_investigation.startdate:
                medata_startdate = (
                    self.icat_investigation.startdate + datetime.timedelta(seconds=1)
                )
            if medata_enddate and medata_enddate >= self.icat_investigation.enddate:
                medata_enddate = self.icat_investigation.enddate - datetime.timedelta(
                    seconds=1
                )

            metadata = {
                "Sample_name": sample_name,
                "startDate": medata_startdate,
                "endDate": medata_enddate,
            }
        else:
            startdate = None
            enddate = None
            metadata = None

        dataset = sync_types.Dataset(
            path=dataset_dir,
            proposal=self.proposal,
            beamline=self.beamline,
            name=dataset,
            raw_root_dir=self.raw_root_dir,
            status_reason=list(status_reason),
            startdate=startdate,
            enddate=enddate,
            metadata=metadata,
        )
        if has_content:
            assert dataset.get_status() == "unregistered"
        else:
            assert dataset.get_status() == "invalid"
        self.add_dataset(dataset)
        return dataset


def _init_unregistered_experiment(
    client: IcatClient, messages, root_dir: str, proposal: str, beamline: str
) -> _ExperimentalSession:
    """Create several datasets on disk:

    - 1 normal dataset starting after the startdate (uregistered)
    - 2 normal datasets (uregistered)
    - 1 dataset with an empty file (uregistered)
    - 2 normal datasets (uregistered)
    - 1 normal dataset ending after the enddate (uregistered)
    """
    startdate = _as_datetime("2023-10-19T08:00:00+02:00")
    enddate = startdate + datetime.timedelta(days=5)

    exp_session = _start_investigation(
        client, messages, root_dir, proposal, beamline, startdate, enddate
    )

    now, dataset_duration, deadtime = _chunk_duration(
        startdate, enddate, 7, deadtime=3600
    )

    # Dataset starts before the startdate
    exp_session.create_dataset(
        "too_soon_collection",
        "0001",
        "too_soon_sample",
        now - 2 * deadtime,
        dataset_duration,
        ["started 1:00:00 before the start of the session"],
    )

    now += dataset_duration + deadtime

    # Normal
    _ = exp_session.create_dataset(
        "first_collection", "0001", "first_sample", now, dataset_duration
    )

    now += dataset_duration + deadtime

    # Normal
    _ = exp_session.create_dataset(
        "first_collection", "0002", "first_sample", now, dataset_duration
    )

    now += dataset_duration + deadtime

    # Empty HDF5 file
    _ = exp_session.create_dataset(
        "empty_file",
        "0001",
        "empty_sample",
        now,
        dataset_duration,
        has_content=False,
        status_reason=[
            "Cannot extract dataset metadata: HDF5 reading error (HDF5 file not created by Bliss)"
        ],
    )

    now += dataset_duration + deadtime

    # Normal
    _ = exp_session.create_dataset(
        "second_collection", "0001", "second_sample", now, dataset_duration
    )

    now += dataset_duration + deadtime

    # Normal
    _ = exp_session.create_dataset(
        "second_collection", "0002", "second_sample", now, dataset_duration
    )

    now += dataset_duration + deadtime

    # Dataset ends after the enddate
    _ = exp_session.create_dataset(
        "too_late_collection",
        "0001",
        "too_late_collection",
        now,
        dataset_duration + 2 * deadtime,
        ["ended 1:00:00 after the end of the session"],
    )

    return exp_session


def _init_empty_files_experiment(
    client: IcatClient, messages, root_dir: str, proposal: str, beamline: str
):
    """Create several datasets on disk:

    - 1 normal dataset (registered)
    - 1 normal dataset (registered without files)
    """
    startdate = _as_datetime("2023-02-01T08:00:00+01:00")
    enddate = startdate + datetime.timedelta(days=5)

    exp_session = _start_investigation(
        client, messages, root_dir, proposal, beamline, startdate, enddate
    )

    now, dataset_duration, deadtime = _chunk_duration(
        startdate, enddate, 2, deadtime=3600
    )

    # Normal
    dataset = exp_session.create_dataset(
        "collection_normal", "0001", "sample_normal", now, dataset_duration
    )
    _store_dataset(client, dataset)
    messages.get(timeout=10)

    now += dataset_duration + deadtime

    # Directory is empty when the file count is calculated (on the ICAT server side)
    dataset = exp_session.create_dataset(
        "collection_missing", "0002", "sample_missing", now, dataset_duration
    )
    with _temporary_empty_directory(dataset.path):
        _store_dataset(client, dataset)
        messages.get(timeout=10)

    return exp_session


@contextmanager
def _temporary_empty_directory(directory_path: str) -> Generator[None, None, None]:
    original_name = path_utils.basename(directory_path)
    new_name = f"{original_name}_tmp"
    renamed_path = os.path.join(path_utils.dirname(directory_path), new_name)
    try:
        os.rename(directory_path, renamed_path)
        os.makedirs(directory_path)
        yield
    finally:
        os.rmdir(directory_path)
        os.rename(renamed_path, directory_path)


def _start_investigation(
    client: IcatClient,
    messages,
    root_dir: str,
    proposal: str,
    beamline: str,
    startdate: datetime.datetime,
    enddate: datetime.datetime,
) -> _ExperimentalSession:
    session = startdate.strftime("%Y%m%d")
    session_dir = path_utils.markdir(
        os.path.join(root_dir, proposal, beamline, session)
    )
    raw_root_dir = path_utils.markdir(os.path.join(session_dir, "RAW_DATA"))
    os.makedirs(raw_root_dir, exist_ok=True)

    exp_session = _ExperimentalSession(
        session_dir=session_dir,
        raw_root_dir=raw_root_dir,
        raw_data_format="esrfv3",
        proposal=proposal,
        beamline=beamline,
        session=session,
        startdate=startdate.date(),
        search_url=f"https://data.esrf.fr/beamline/{beamline}?search={proposal}",
        datasets=dict(),
    )

    client.start_investigation(
        beamline=beamline,
        proposal=proposal,
        start_datetime=startdate,
        end_datetime=enddate,
    )
    messages.get(timeout=10)

    exp_session.icat_investigation = sync_types.IcatInvestigation(
        id=0,
        url="https://data.esrf.fr/investigation/0/datasets",
        search_url=f"https://data.esrf.fr/beamline/{beamline}?search={proposal}",
        registered_datasets=list(),
        startdate=startdate,
        enddate=enddate,
    )

    return exp_session


def _store_dataset(client: IcatClient, dataset: sync_types.Dataset) -> None:
    client.store_dataset(
        beamline=dataset.beamline,
        proposal=dataset.proposal,
        dataset=dataset.name,
        path=dataset.path,
        metadata=dataset.metadata,
    )


def _chunk_duration(
    startdate: datetime.datetime,
    enddate: datetime.datetime,
    nchunks: int,
    deadtime: int = 0,
):
    deadtime = datetime.timedelta(seconds=deadtime)
    total_deadtime = (nchunks + 1) * deadtime
    total_time = enddate - startdate
    chunk_duration = (total_time - total_deadtime) / nchunks
    now = startdate + deadtime
    return now, chunk_duration, deadtime


def _save_scan(
    nxroot: h5py.File,
    startdate: datetime.datetime,
    duration: datetime.timedelta,
    sample_name: str,
    failed: bool = False,
) -> Optional[datetime.datetime]:
    scans = list(nxroot)
    if scans:
        scan = max(map(int, map(float, scans))) + 1
    else:
        scan = 1
    name = f"{scan}.1"
    grp = nxroot.create_group(name)

    grp["start_time"] = startdate.isoformat()
    if failed:
        return

    enddate = startdate + duration
    grp["end_time"] = enddate.isoformat()
    grp["sample/name"] = sample_name
    return enddate


@contextmanager
def _init_dataset(
    raw_root_dir: str,
    collection: str,
    dataset: str,
) -> Generator[h5py.File, None, None]:
    basename = f"{collection}_{dataset}"
    dataset_dir = os.path.join(raw_root_dir, collection, basename)
    os.makedirs(dataset_dir, exist_ok=True)

    filename = os.path.basename(dataset_dir) + ".h5"
    with h5py.File(os.path.join(dataset_dir, filename), mode="w") as f:
        yield f


def _as_datetime(isoformat: str) -> datetime.datetime:
    return datetime.datetime.fromisoformat(isoformat).astimezone()


def _astimezone_string(tzinfo, iso: bool = True) -> Callable[[datetime.datetime], str]:
    def astimezone_string(dt: datetime.datetime) -> str:
        dt = dt.astimezone(tz=tzinfo)
        if iso:
            return dt.isoformat()
        return str(dt)

    return astimezone_string



import sys
import logging
import argparse
import datetime
from typing import Mapping, Union, Optional

import h5py
import numpy

from ..utils.log_utils import basic_config
from ..metadata.nexus import create_nxtreedict
from ..metadata.definitions import load_icat_fields

logger = logging.getLogger(__name__)


def save_icat_as_nexus(filename: str, url: Optional[str] = None) -> None:
    icat_fields = load_icat_fields(url=url)

    metadict = _example_icat_metadata_as_dict(icat_fields)
    nxtreedict = create_nxtreedict(
        metadict, icat_fields=icat_fields, add_icat_attrs=True
    )

    with h5py.File(filename, "w") as nxroot:
        nxroot.attrs["NX_class"] = "NXroot"
        _dicttonx("entryname", nxtreedict, nxroot)


def _example_icat_metadata_as_dict(icat_fields) -> dict:
    """Example ICAT field values are generated based on the data type."""
    now = datetime.datetime.now()

    values = {
        "NX_CHAR": lambda name: name,
        "NX_DATE_TIME": lambda name: (
            now + datetime.timedelta(minutes=len(name))
        ).isoformat(),
        "NX_FLOAT64": lambda name: numpy.float64(len(name) + 0.001),
        "NX_FLOAT": lambda name: numpy.float32(len(name) + 0.1),
        "NX_BOOLEAN": lambda name: True if len(name) % 2 == 0 else False,
        "NX_INT": lambda name: numpy.int32(len(name) + 1),
    }

    metadict = dict()
    for field in icat_fields.iter_fields():
        value_gen = values[field.nxtype]
        field_name = field.field_name
        metadict[field_name] = value_gen(field_name)
    return metadict


def _dicttonx(
    name: str, value: Union[Mapping, str, float], h5group: h5py.Group
) -> None:
    if "@" in name:
        assert isinstance(value, str), f"Attribute {name} must be a string"
        dsetname, _, attrname = name.rpartition("@")
        if dsetname:
            h5item = h5group[dsetname]
        else:
            h5item = h5group
        h5item.attrs[attrname] = value
        return
    assert "@" not in name, f"Item {name} cannot have an '@' character"

    if isinstance(value, Mapping):
        h5group = h5group.create_group(name)
        for name, value in value.items():
            _dicttonx(name, value, h5group)
    else:
        h5group[name] = value


def main(argv=None):
    basic_config(logger=logger, level=logging.DEBUG, format="%(message)s")

    if argv is None:
        argv = sys.argv

    parser = argparse.ArgumentParser(
        description="Save ICAT definitions in a NeXus compliant HDF5 file"
    )
    parser.add_argument(
        "--filename",
        type=str,
        required=False,
        default="icat.h5",
        help="File name to store the ICAT definitions",
    )
    parser.add_argument(
        "--url", type=str, required=False, help="URL of the ICAT definitions"
    )

    args = parser.parse_args(argv[1:])

    save_icat_as_nexus(args.filename, args.url)
    logger.info(f"NeXus compliant HDF5 file created: '{args.filename}'")


if __name__ == "__main__":
    sys.exit(main())



import datetime
import warnings
from typing import List, Mapping, Optional, Sequence, Tuple, Union

import numpy

from . import defaults
from .archive import IcatArchiveStatusClient, StatusLevel, StatusType
from .elogbook import IcatElogbookClient
from .interface import DatasetId, Dataset, IcatClientInterface
from .investigation import IcatInvestigationClient
from .metadata import IcatMetadataClient
from .update_metadata import IcatUpdateMetadataClient
from .add_files import IcatAddFilesClient
from .icatplus_restricted import IcatPlusRestrictedClient


class IcatClient(IcatClientInterface):
    """Client object that provides access to these services:

    - ActiveMQ message broker for creating datasets in ICAT
    - ActiveMQ message broker for updating dataset metadata in ICAT
    - ActiveMQ message broker for updating dataset file count in ICAT
    - ActiveMQ message broker for updating dataset archiving status in ICAT
    - RESTful interface for sending electronic logbook messages/images and get information about investigations

    The RESTful interface is referred to as ICAT+ and the ActiveMQ message brokers are consumed by the "ingesters".
    """

    def __init__(
        self,
        metadata_urls: Optional[List[str]] = None,
        elogbook_url: Optional[str] = None,
        elogbook_token: Optional[str] = None,
        metadata_queue: Optional[str] = None,
        metadata_queue_monitor_port: Optional[int] = None,
        elogbook_timeout: Optional[float] = None,
        feedback_timeout: Optional[float] = None,
        queue_timeout: Optional[float] = None,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        elogbook_metadata: Optional[Mapping] = None,
        archive_urls: Optional[List[str]] = None,
        archive_queue: Optional[str] = None,
        archive_queue_monitor_port: Optional[int] = None,
        update_metadata_urls: Optional[List[str]] = None,
        update_metadata_queue: Optional[str] = None,
        update_metadata_queue_monitor_port: Optional[int] = None,
        add_files_urls: Optional[List[str]] = None,
        add_files_queue: Optional[str] = None,
        add_files_queue_monitor_port: Optional[int] = None,
        reschedule_investigation_urls: Optional[List[str]] = None,
        reschedule_investigation_queue: Optional[str] = None,
        reschedule_investigation_queue_monitor_port: Optional[int] = None,
        icatplus_restricted_url: Optional[str] = None,
        icatplus_password: Optional[str] = None,
        icat_session_id: Optional[str] = None,
        catalogue_queues: Optional[List[str]] = None,  # DEPRECATED
        catalogue_url: Optional[str] = None,  # DEPRECATED
        tracking_url: Optional[str] = None,  # DEPRECATED
    ):
        """
        :param metadata_urls: URLs of the ActiveMQ message brokers to be used for creating ICAT datasets from a directory with metadata.
        :param elogbook_url: URL of the ICAT+ REST server to be used for sending text or images to the electronic logbook and get information about investigations.
        :param elogbook_token: Access token for restricted requests to `elogbook_url`.
        :param metadata_queue: Queue to be used when sending a message to `metadata_urls`.
        :param metadata_queue_monitor_port: REST server port to be used for monitor the `metadata_urls` ActiveMQ message brokers (same host as the message broker).
        :param elogbook_timeout: POST timeout for `elogbook_url`.
        :param feedback_timeout: GET timeout for `elogbook_url`.
        :param queue_timeout: Connection timeout for the ActiveMQ message brokers.
        :param beamline: Default beamline to be used as metadata when sending messages to `metadata_urls` or `elogbook_url`.
        :param proposal: Default proposal to be used as metadata when sending messages to `metadata_urls` or `elogbook_url`.
        :param elogbook_metadata: Default electronic logbook metadata to be used when sending messages to  `elogbook_url`.
        :param archive_urls: URLs of the ActiveMQ message brokers to be used for updating the archival status of ICAT datasets.
        :param archive_queue: Queue to be used when sending a message to `archive_urls`.
        :param archive_queue_monitor_port: REST server port to be used for monitor the `archive_urls` ActiveMQ message brokers (same host as the message broker).
        :param update_metadata_urls: URLs of the ActiveMQ message brokers to be used for update metadata of ICAT datasets.
        :param update_metadata_queue: Queue to be used when sending a message to `update_metadata_urls`.
        :param update_metadata_queue_monitor_port: REST server port to be used for monitor the `update_metadata_urls` ActiveMQ message brokers (same host as the message broker).
        :param add_files_urls: URLs of the ActiveMQ message brokers to be used for updating the file count of ICAT datasets.
        :param add_files_queue: Queue to be used when sending a message to `add_files_urls`.
        :param add_files_queue_monitor_port: REST server port to be used for monitor the `add_files_urls` ActiveMQ message brokers (same host as the message broker).
        :param reschedule_investigation_urls: URLs of the ActiveMQ message brokers to be used for rescheduling investigations.
        :param reschedule_investigation_queue: Queue to be used when sending a message to `reschedule_investigation`.
        :param reschedule_investigation_queue_monitor_port: REST server port to be used for monitor the `reschedule_investigation` ActiveMQ message brokers (same host as the message broker).
        :param icatplus_restricted_url: URL of the ICAT+ REST server to be used for restricted access (requires `icatplus_password` or `do_log_in`).
        :param icatplus_password: Password to provide access to `icatplus_restricted_url`.
        :param icat_session_id: The ICAT's session id that is used for authentication/authorization in ICAT.
        :param catalogue_queues: URLs of the ActiveMQ message brokers to be used for the catalogue (DEPRECATED).
        :param catalogue_url: URL of the ICAT+ REST server to be used for accessing the catalogue (DEPRECATED).
        :param tracking_url: URL of the ICAT+ REST server to be used for accessing the tracking (DEPRECATED).
        """

        # Defaults to be used in client methods
        self.current_proposal = proposal
        self.current_beamline = beamline
        self.current_dataset = None
        self.current_path = None
        self.current_dataset_metadata = None

        # Deprecated constructor arguments
        if catalogue_queues:
            reschedule_investigation_urls = catalogue_queues
            warnings.warn(
                "'catalogue_queues' is deprecated and will be renamed to 'reschedule_investigation_urls'.",
                DeprecationWarning,
            )
        if catalogue_url:
            warnings.warn(
                "'catalogue_url' is deprecated and will be renamed to 'icatplus_restricted_url'.",
                DeprecationWarning,
            )
            if not icatplus_restricted_url:
                icatplus_restricted_url = catalogue_url
        if tracking_url:
            warnings.warn(
                "'tracking_url' is deprecated and will be renamed to 'icatplus_restricted_url'.",
                DeprecationWarning,
            )
            if not icatplus_restricted_url:
                icatplus_restricted_url = tracking_url

        # Initialize clients to message brokers or REST URL's.
        # Clients with missing parameters are not initialized
        # add the associated client methods will raise an error.
        self._init_metadata_client(
            brokers=metadata_urls,
            queue=metadata_queue,
            timeout=queue_timeout,
            monitor_port=metadata_queue_monitor_port,
        )
        self._init_archive_client(
            brokers=archive_urls,
            queue=archive_queue,
            timeout=queue_timeout,
            monitor_port=archive_queue_monitor_port,
        )
        self._init_update_metadata_client(
            brokers=update_metadata_urls,
            queue=update_metadata_queue,
            timeout=queue_timeout,
            monitor_port=update_metadata_queue_monitor_port,
        )
        self._init_elogbook_client(
            url=elogbook_url,
            api_key=elogbook_token,
            timeout=elogbook_timeout,
            metadata=elogbook_metadata,
        )
        self._init_investigation_client(
            url=elogbook_url,
            api_key=elogbook_token,
            timeout=feedback_timeout,
        )
        self._init_add_files_client(
            brokers=add_files_urls,
            queue=add_files_queue,
            timeout=queue_timeout,
            monitor_port=add_files_queue_monitor_port,
        )
        self._init_reschedule_investigation_client(
            brokers=reschedule_investigation_urls,
            queue=reschedule_investigation_queue,
            timeout=queue_timeout,
            monitor_port=reschedule_investigation_queue_monitor_port,
        )
        self._init_icatplus_restricted_client(
            url=icatplus_restricted_url,
            password=icatplus_password,
            session_id=icat_session_id,
        )

    def disconnect(self):
        if self.__metadata_client is not None:
            self.__metadata_client.disconnect()
        if self.__update_metadata_client is not None:
            self.__update_metadata_client.disconnect()
        if self.__add_files_client is not None:
            self.__add_files_client.disconnect()
        if self.__archive_client is not None:
            self.__archive_client.disconnect()
        if self.__reschedule_investigation_client is not None:
            self.__reschedule_investigation_client.disconnect()

    @property
    def metadata_client(self):
        warnings.warn(
            "Will be removed in the next release.", DeprecationWarning, stacklevel=2
        )
        return self._metadata_client

    @property
    def _metadata_client(self):
        if self.__metadata_client is None:
            raise RuntimeError("The message queue URL's are missing")
        return self.__metadata_client

    def _init_metadata_client(
        self,
        brokers: Optional[List[str]] = None,
        queue: Optional[str] = None,
        timeout: Optional[float] = None,
        monitor_port: Optional[int] = None,
    ):
        if brokers:
            self.__metadata_client = IcatMetadataClient(
                queue_urls=brokers,
                queue_name=queue,
                monitor_port=monitor_port,
                timeout=timeout,
            )
        else:
            self.__metadata_client = None

    @property
    def elogbook_client(self):
        warnings.warn(
            "Will be removed in the next release.", DeprecationWarning, stacklevel=2
        )
        return self._elogbook_client

    @property
    def _elogbook_client(self):
        if self.__elogbook_client is None:
            raise RuntimeError("The ICAT+ URL and/or token are missing")
        return self.__elogbook_client

    def _init_elogbook_client(
        self,
        url: Optional[str] = None,
        api_key: Optional[str] = None,
        timeout: Optional[float] = None,
        metadata: Optional[dict] = None,
    ):
        if url and api_key:
            if metadata is None:
                metadata = dict()
            self.__elogbook_client = IcatElogbookClient(
                url=url,
                api_key=api_key,
                timeout=timeout,
                **metadata,
            )
        else:
            self.__elogbook_client = None

    @property
    def investigation_client(self):
        warnings.warn(
            "Will be removed in the next release.", DeprecationWarning, stacklevel=2
        )
        return self._investigation_client

    @property
    def _investigation_client(self):
        if self.__investigation_client is None:
            raise RuntimeError("The ICAT+ URL and/or token are missing")
        return self.__investigation_client

    def _init_investigation_client(
        self,
        url: Optional[str] = None,
        api_key: Optional[str] = None,
        timeout: Optional[float] = None,
    ):
        if url and api_key:
            self.__investigation_client = IcatInvestigationClient(
                url=url, api_key=api_key, timeout=timeout
            )
        else:
            self.__investigation_client = None

    @property
    def archive_client(self):
        warnings.warn(
            "Will be removed in the next release.", DeprecationWarning, stacklevel=2
        )
        return self._archive_client

    @property
    def _archive_client(self):
        if self.__archive_client is None:
            raise RuntimeError("The message queue URL's are missing")
        return self.__archive_client

    def _init_archive_client(
        self,
        brokers: Optional[List[str]] = None,
        queue: Optional[str] = None,
        timeout: Optional[float] = None,
        monitor_port: Optional[int] = None,
    ):
        if brokers:
            self.__archive_client = IcatArchiveStatusClient(
                queue_urls=brokers,
                queue_name=queue,
                monitor_port=monitor_port,
                timeout=timeout,
            )
        else:
            self.__archive_client = None

    @property
    def update_metadata_client(self):
        warnings.warn(
            "Will be removed in the next release.", DeprecationWarning, stacklevel=2
        )

    @property
    def _update_metadata_client(self):
        if self.__update_metadata_client is None:
            raise RuntimeError("The message queue URL's are missing")
        return self.__update_metadata_client

    def _init_update_metadata_client(
        self,
        brokers: Optional[List[str]] = None,
        queue: Optional[str] = None,
        timeout: Optional[float] = None,
        monitor_port: Optional[int] = None,
    ):
        if brokers:
            self.__update_metadata_client = IcatUpdateMetadataClient(
                queue_urls=brokers,
                queue_name=queue,
                monitor_port=monitor_port,
                timeout=timeout,
            )
        else:
            self.__update_metadata_client = None

    @property
    def add_files_client(self):
        warnings.warn(
            "Will be removed in the next release.", DeprecationWarning, stacklevel=2
        )
        return self._add_files_client

    @property
    def _add_files_client(self):
        if self.__add_files_client is None:
            raise RuntimeError("The message queue URL's are missing")
        return self.__add_files_client

    def _init_add_files_client(
        self,
        brokers: Optional[List[str]] = None,
        queue: Optional[str] = None,
        timeout: Optional[float] = None,
        monitor_port: Optional[int] = None,
    ):
        if brokers:
            self.__add_files_client = IcatAddFilesClient(
                queue_urls=brokers,
                queue_name=queue,
                monitor_port=monitor_port,
                timeout=timeout,
            )
        else:
            self.__add_files_client = None

    @property
    def catalogue_client(self):
        warnings.warn(
            "Will be removed in the next release.", DeprecationWarning, stacklevel=2
        )
        return self

    @property
    def tracking_client(self):
        warnings.warn(
            "Will be removed in the next release.", DeprecationWarning, stacklevel=2
        )
        return self

    @property
    def _icatplus_authentication_client(self):
        if self.__icatplus_authentication_client is None:
            raise RuntimeError("The ICAT+ URL and/or token are missing")
        return self.__icatplus_authentication_client

    @property
    def _icatplus_restricted_client(self):
        if self.__icatplus_restricted_client is None:
            raise RuntimeError("The ICAT+ URL is missing and/or login")
        return self.__icatplus_restricted_client

    def _init_icatplus_restricted_client(
        self,
        url: Optional[str] = None,
        password: Optional[str] = None,
        session_id: Optional[str] = None,
    ):
        if url:
            self.__icatplus_restricted_client = IcatPlusRestrictedClient(
                url=url, password=password, session_id=session_id
            )
        else:
            self.__icatplus_restricted_client = None

    @property
    def _reschedule_investigation_client(self):
        if self.__reschedule_investigation_client is None:
            raise RuntimeError("The message queue URL's are missing")
        return self.__reschedule_investigation_client

    def _init_reschedule_investigation_client(
        self,
        brokers: Optional[List[str]] = None,
        queue: Optional[str] = None,
        timeout: Optional[float] = None,
        monitor_port: Optional[int] = None,
    ):
        if brokers:
            self.__reschedule_investigation_client = IcatMetadataClient(
                queue_urls=brokers or defaults.RESCHEDULE_INVESTIGATION_BROKERS,
                queue_name=queue or defaults.RESCHEDULE_INVESTIGATION_QUEUE,
                monitor_port=monitor_port,
                timeout=timeout,
            )
        else:
            self.__reschedule_investigation_client = None

    @property
    def current_proposal(self):
        return self.__current_proposal

    @current_proposal.setter
    def current_proposal(self, value: Optional[str]):
        self.__current_proposal = value

    @property
    def current_beamline(self):
        return self.__current_beamline

    @current_beamline.setter
    def current_beamline(self, value: Optional[str]):
        self.__current_beamline = value

    @property
    def current_dataset(self):
        return self.__current_dataset

    @current_dataset.setter
    def current_dataset(self, value: Optional[str]):
        self.__current_dataset = value

    @property
    def current_dataset_metadata(self):
        return self.__current_dataset_metadata

    @current_dataset_metadata.setter
    def current_dataset_metadata(self, value: Optional[dict]):
        self.__current_dataset_metadata = value

    @property
    def current_path(self):
        return self.__current_path

    @current_path.setter
    def current_path(self, value: Optional[str]):
        self.__current_path = value

    def send_message(
        self,
        msg: str,
        msg_type: Optional[str] = None,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        dataset: Optional[str] = None,
        beamline_only: Optional[bool] = None,
        editable: Optional[bool] = None,
        formatted: Optional[bool] = None,
        mimetype: Optional[str] = None,
        **payload,
    ):
        if beamline_only:
            proposal = None
        elif proposal is None:
            proposal = self.current_proposal
        if beamline is None:
            beamline = self.current_beamline
        if beamline_only:
            dataset = None
        elif dataset is None:
            dataset = self.current_dataset
        self._elogbook_client.send_message(
            message=msg,
            message_type=msg_type,
            beamline=beamline,
            proposal=proposal,
            dataset=dataset,
            editable=editable,
            formatted=formatted,
            mimetype=mimetype,
            **payload,
        )

    def send_binary_data(
        self,
        data: bytes,
        mimetype: Optional[str] = None,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        beamline_only: Optional[bool] = None,
        **payload,
    ):
        if beamline_only:
            proposal = None
        elif proposal is None:
            proposal = self.current_proposal
        if beamline is None:
            beamline = self.current_beamline
        self._elogbook_client.send_binary_data(
            data, mimetype=mimetype, beamline=beamline, proposal=proposal, **payload
        )

    def send_text_file(
        self,
        filename: str,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        dataset: Optional[str] = None,
        beamline_only: Optional[bool] = None,
        **payload,
    ):
        if beamline_only:
            proposal = None
        elif proposal is None:
            proposal = self.current_proposal
        if beamline is None:
            beamline = self.current_beamline
        if beamline_only:
            dataset = None
        elif dataset is None:
            dataset = self.current_dataset
        self._elogbook_client.send_text_file(
            filename, beamline=beamline, proposal=proposal, dataset=dataset, **payload
        )

    def send_binary_file(
        self,
        filename: str,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        beamline_only: Optional[bool] = None,
        **payload,
    ):
        if beamline_only:
            proposal = None
        elif proposal is None:
            proposal = self.current_proposal
        if beamline is None:
            beamline = self.current_beamline
        self._elogbook_client.send_binary_file(
            filename, beamline=beamline, proposal=proposal, **payload
        )

    def start_investigation(
        self,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        start_datetime=None,
        end_datetime=None,
    ):
        if proposal is None:
            proposal = self.current_proposal
        else:
            self.current_proposal = proposal
        if beamline is None:
            beamline = self.current_beamline
        else:
            self.current_beamline = beamline
        self._metadata_client.start_investigation(
            beamline=beamline,
            proposal=proposal,
            start_datetime=start_datetime,
            end_datetime=end_datetime,
        )

    def store_dataset(
        self,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        dataset: Optional[str] = None,
        path: Optional[str] = None,
        metadata: dict = None,
        store_filename: Optional[str] = None,
    ):
        if proposal is None:
            proposal = self.current_proposal
        if beamline is None:
            beamline = self.current_beamline
        if dataset is None:
            dataset = self.current_dataset
        if path is None:
            path = self.current_path
        if metadata is None:
            metadata = self.current_dataset_metadata
            if metadata is None:
                metadata = dict()
        if store_filename:
            self._metadata_client.store_metadata(
                store_filename,
                beamline=beamline,
                proposal=proposal,
                dataset=dataset,
                path=path,
                metadata=metadata,
            )
        else:
            self._metadata_client.send_metadata(
                beamline=beamline,
                proposal=proposal,
                dataset=dataset,
                path=path,
                metadata=metadata,
            )

    def store_processed_data(
        self,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        dataset: Optional[str] = None,
        path: Optional[str] = None,
        metadata: dict = None,
        raw: Sequence = tuple(),
        store_filename: Optional[str] = None,
    ):
        """The 'raw' argument is shorthand for `metadata = {'input_datasets': ...}`."""
        if metadata is None:
            metadata = self.current_dataset_metadata
            if metadata is None:
                metadata = dict()
        if raw:
            if isinstance(raw, str):
                metadata["input_datasets"] = [raw]
            elif isinstance(raw, Sequence):
                metadata["input_datasets"] = list(raw)
            else:
                metadata["input_datasets"] = [raw]
        if not metadata.get("input_datasets"):
            raise ValueError("Provide 'raw' dataset directories")
        self.store_dataset(
            beamline=beamline,
            proposal=proposal,
            dataset=dataset,
            path=path,
            metadata=metadata,
            store_filename=store_filename,
        )

    def store_dataset_from_file(self, store_filename: Optional[str] = None):
        self._metadata_client.send_metadata_from_file(store_filename)

    def investigation_info(
        self,
        beamline: str,
        proposal: str,
        date: Optional[Union[datetime.datetime, datetime.date]] = None,
        allow_open_ended: bool = True,
        timeout: Optional[float] = None,
    ) -> Optional[dict]:
        return self._investigation_client.investigation_info(
            beamline=beamline,
            proposal=proposal,
            date=date,
            allow_open_ended=allow_open_ended,
            timeout=timeout,
        )

    def registered_dataset_ids(
        self,
        beamline: str,
        proposal: str,
        date: Optional[Union[datetime.datetime, datetime.date]] = None,
        allow_open_ended: bool = True,
        timeout: Optional[float] = None,
    ) -> Optional[List[DatasetId]]:
        return self._investigation_client.registered_dataset_ids(
            beamline=beamline,
            proposal=proposal,
            date=date,
            allow_open_ended=allow_open_ended,
            timeout=timeout,
        )

    def registered_datasets(
        self,
        beamline: str,
        proposal: str,
        date: Optional[Union[datetime.datetime, datetime.date]] = None,
        allow_open_ended: bool = True,
        timeout: Optional[float] = None,
    ) -> Optional[List[Dataset]]:
        return self._investigation_client.registered_datasets(
            beamline=beamline,
            proposal=proposal,
            date=date,
            allow_open_ended=allow_open_ended,
            timeout=timeout,
        )

    def investigation_info_string(
        self,
        beamline: str,
        proposal: str,
        date: Optional[Union[datetime.datetime, datetime.date]] = None,
        allow_open_ended: bool = True,
        timeout: Optional[float] = None,
    ) -> str:
        info = self.investigation_info(
            beamline=beamline,
            proposal=proposal,
            date=date,
            allow_open_ended=allow_open_ended,
            timeout=timeout,
        )
        if info:
            rows = [(str(k), str(v)) for k, v in info.items()]
            lengths = numpy.array([[len(s) for s in row] for row in rows])
            fmt = "   ".join(["{{:<{}}}".format(n) for n in lengths.max(axis=0)])
            infostr = "ICAT proposal time slot:\n "
            infostr += "\n ".join([fmt.format(*row) for row in rows])
        elif info is None:
            infostr = f"Proposal information currently not available ({self.reason_for_missing_information})"
        else:
            infostr = "Proposal NOT available in the data portal"
        return infostr

    def investigation_summary(
        self,
        beamline: str,
        proposal: str,
        date: Optional[Union[datetime.datetime, datetime.date]] = None,
        allow_open_ended: bool = True,
        timeout: Optional[float] = None,
    ) -> List[Tuple]:
        info = self.investigation_info(
            beamline=beamline,
            proposal=proposal,
            date=date,
            allow_open_ended=allow_open_ended,
            timeout=timeout,
        )
        keys = ["e-logbook", "data portal"]
        if info:
            rows = [(key, info[key]) for key in keys]
        elif info is None:
            rows = [
                (
                    key,
                    f"Proposal information currently not available ({self.reason_for_missing_information})",
                )
                for key in keys
            ]
        else:
            rows = [(key, "Proposal NOT available in the data portal") for key in keys]
        return rows

    def update_archive_restore_status(
        self,
        dataset_id: int = None,
        type: StatusType = None,
        level: StatusLevel = StatusLevel.INFO,
        message: Optional[str] = None,
    ):
        self._archive_client.send_archive_status(
            dataset_id=dataset_id, type=type, level=level, message=message
        )

    def update_metadata(
        self,
        proposal: str = None,
        beamline: str = None,
        dataset_paths: str = None,
        metadata_name: str = None,
        metadata_value: str = None,
    ):
        if proposal is None:
            proposal = self.current_proposal
        if beamline is None:
            beamline = self.current_beamline
        self.__update_metadata_client.send_update_metadata(
            proposal=proposal,
            beamline=beamline,
            dataset_paths=dataset_paths,
            metadata_name=metadata_name,
            metadata_value=metadata_value,
        )

    def add_files(
        self,
        dataset_id: int = None,
    ):
        self.__add_files_client.add_files(
            dataset_id=dataset_id,
        )

    def reschedule_investigation(self, investigation_id: str):
        self._reschedule_investigation_client.reschedule_investigation(investigation_id)

    def do_log_in(self, password: str) -> dict:
        return self._icatplus_restricted_client.login(password)

    def get_investigations_by(
        self,
        filter: Optional[str] = None,
        instrument_name: Optional[str] = None,
        start_date: Optional[datetime.datetime] = None,
        end_date: Optional[datetime.datetime] = None,
        ids: Optional[str] = None,
    ) -> List[dict]:
        return self._icatplus_restricted_client.get_investigations_by(
            filter=filter,
            instrument_name=instrument_name,
            start_date=start_date,
            end_date=end_date,
            ids=ids,
        )

    def get_samples_by(
        self,
        investigationId: Optional[str] = None,
        sampleIds: Optional[str] = None,
    ) -> List[dict]:
        return self._icatplus_restricted_client.get_samples_by(
            investigationId=investigationId, sampleIds=sampleIds
        )

    def get_parcels_by(self, investigation_id: str) -> List[dict]:
        return self._icatplus_restricted_client.get_parcels_by(investigation_id)

    def get_session_information(self) -> dict:
        return self._icatplus_restricted_client.get_session_information()

    @property
    def expire_datasets_on_close(self) -> bool:
        return False

    @property
    def reason_for_missing_information(self) -> str:
        return "ICAT communication timeout"



import base64
import socket
import logging
import datetime
import mimetypes
import warnings
from enum import Enum
from urllib.parse import urljoin
from typing import Optional, Iterable, List, Tuple

import requests

from ..utils.url import normalize_url
from . import defaults
from importlib.metadata import version as get_version

release = get_version("pyicat_plus")

logger = logging.getLogger(__name__)


class IcatElogbookClient:
    """Client for the e-logbook part of the ICAT+ REST API.

    REST API docs:
    https://icatplus.esrf.fr/api-docs/

    The ICAT+ server project:
    https://gitlab.esrf.fr/icat/icat-plus/-/blob/master/README.md
    """

    DEFAULT_SCHEME = "https"

    def __init__(
        self,
        url: str,
        api_key: Optional[str] = None,
        timeout: Optional[float] = None,
        **payload,
    ):
        if api_key is None:
            api_key = defaults.ELOGBOOK_TOKEN
        url = normalize_url(url, default_scheme=self.DEFAULT_SCHEME)

        path = f"dataacquisition/{api_key}/notification"
        self._message_url = urljoin(url, path)

        path = f"dataacquisition/{api_key}/base64"
        self._data_url = urljoin(url, path)

        self._init_payload = payload
        self._init_payload.setdefault("machine", socket.getfqdn())
        self._init_payload.setdefault("software", "pyicat-plus_v" + release)

        self.raise_error = True
        if timeout is None:
            timeout = 0.1
        self.timeout = timeout

    def _merge_payloads(self, message_payload: dict, call_payload: dict) -> dict:
        payloads = self._sorted_payloads(message_payload, call_payload)
        result = {k: v for payload in payloads for k, v in payload.items()}
        tags = self._merge_payload_tags(*payloads)
        if tags:
            result.pop("tags", None)
            result["tag"] = tags
        return result

    def _sorted_payloads(self, message_payload: dict, call_payload: dict) -> List[dict]:
        """Sorted by increasing priority"""
        return [message_payload, self._init_payload, call_payload]

    def _merge_payload_tags(self, *payloads: Iterable[dict]) -> List[dict]:
        """The payload tags can be eithers a list of strings or a list of dictionaries.
        The return value are the merged tags as a list of dictionaries.
        """
        names = set()
        tags = list()
        for payload in payloads:
            ptags = payload.get("tag", list()) + payload.get("tags", list())
            for tag in ptags:
                if isinstance(tag, str):
                    if tag in names:
                        continue
                    names.add(tag)
                    tags.append({"name": tag})
                else:
                    if tag["name"] in names:
                        continue
                    names.add(tag["name"])
                    tags.append(tag)
        return tags

    def _post_with_payload(
        self, url: str, message_payload: dict, call_payload: dict
    ) -> None:
        payload = self._merge_payloads(message_payload, call_payload)
        payload.setdefault(
            "creationDate", datetime.datetime.now().astimezone().isoformat()
        )
        try:
            response = requests.post(url, json=payload, timeout=self.timeout)
        except requests.exceptions.ReadTimeout:
            return  # we have no confirmation that the call succeeded
        except Exception as e:
            if self.raise_error:
                raise
            logger.exception(e)
            return
        if self.raise_error:
            response.raise_for_status()
        elif not response.ok:
            logger.error("%s: %s", response, response.text)

    def send_message(
        self,
        message: str,
        message_type: Optional[str] = None,
        editable: Optional[bool] = None,
        formatted: Optional[bool] = None,
        mimetype: Optional[str] = None,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        dataset: Optional[str] = None,
        **call_payload,
    ):
        url = self._compose_url(
            url=self._message_url, beamline=beamline, proposal=proposal
        )
        message_payload = self._encode_message(
            message,
            message_type=message_type,
            editable=editable,
            formatted=formatted,
            mimetype=mimetype,
            dataset=dataset,
        )
        self._post_with_payload(url, message_payload, call_payload)

    def send_binary_data(
        self,
        data: bytes,
        mimetype: str,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        **call_payload,
    ):
        url = self._compose_url(
            url=self._data_url, beamline=beamline, proposal=proposal
        )
        message_payload = self._encode_binary_data(data, mimetype=mimetype)
        self._post_with_payload(url, message_payload, call_payload)

    @staticmethod
    def _compose_url(
        url: str, beamline: Optional[str] = None, proposal: Optional[str] = None
    ):
        query = {}
        if beamline:
            query["instrumentName"] = beamline
        if proposal:
            query["investigationName"] = proposal
        query = "&".join([f"{k}={v}" for k, v in query.items()])
        return f"{url}?{query}"

    def send_text_file(
        self,
        filename: str,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        dataset: Optional[str] = None,
        message_type: Optional[str] = None,
        editable: Optional[bool] = None,
        formatted: Optional[bool] = None,
        mimetype: Optional[str] = None,
        **payload,
    ):
        with open(filename, "r") as f:
            message = f.read()
        self.send_message(
            message,
            message_type=message_type,
            proposal=proposal,
            beamline=beamline,
            dataset=dataset,
            editable=editable,
            formatted=formatted,
            mimetype=mimetype,
            **payload,
        )

    def send_binary_file(
        self,
        filename: str,
        beamline: Optional[str] = None,
        proposal: Optional[str] = None,
        **payload,
    ):
        with open(filename, "rb") as f:
            data = f.read()
        mimetype, _ = mimetypes.guess_type(filename, strict=True)
        self.send_binary_data(
            data, mimetype=mimetype, beamline=beamline, proposal=proposal, **payload
        )

    def _encode_message(
        self,
        message: str,
        message_type: Optional[str] = None,
        editable: Optional[bool] = None,
        formatted: Optional[bool] = None,
        mimetype: Optional[str] = None,
        dataset: Optional[str] = None,
    ) -> dict:
        message_category, message_type = _message_category_and_type(
            message_type=message_type, editable=editable, formatted=formatted
        )
        if mimetype is None:
            mimetype = "text/plain"
        try:
            format = _MessageFormatMapping[mimetype]
        except KeyError:
            raise ValueError(
                f"mime type '{mimetype}' is not supported ({list(_MessageFormatMapping)})"
            ) from None
        message = {
            "type": message_type.name,
            "category": message_category.name,
            "content": [{"format": format.name, "text": message}],
        }
        if dataset:
            message["datasetName"] = dataset
        return message

    def _encode_binary_data(
        self,
        data: bytes,
        mimetype: Optional[str] = None,
    ) -> dict:
        if not mimetype:
            # arbitrary binary data
            mimetype = "application/octet-stream"
        data_header = f"data:{mimetype};base64,"
        data_blob = base64.b64encode(data).decode("latin-1")
        return {"base64": data_header + data_blob}


_MessageCategory = Enum("MessageCategory", "debug info error commandLine comment")
_MessageType = Enum("MessageType", "annotation notification")
_MessageFormat = Enum("MessageType", "plainText html")

_MessageCategoryMapping = {
    "debug": _MessageCategory.debug,
    "info": _MessageCategory.info,
    "warning": _MessageCategory.error,
    "warn": _MessageCategory.error,
    "error": _MessageCategory.error,
    "critical": _MessageCategory.error,
    "fatal": _MessageCategory.error,
    "command": _MessageCategory.commandLine,
    "comment": _MessageCategory.comment,
}

_MessageFormatMapping = {
    "text/plain": _MessageFormat.plainText,
    "text/html": _MessageFormat.html,
}


def _message_category_and_type(
    message_type: Optional[str] = None,
    editable: Optional[bool] = None,
    formatted: Optional[bool] = None,
) -> Tuple[_MessageCategory, _MessageType]:
    """Derive the ICAT message category from the API message type.

    The ICAT message types are:

    - "annotation" (default for comments): editable and unformatted message
    - "notification" (default for non-comments): formatted and not editable

    Only comments can be editable and a message cannot be editable and
    formatted at the same time.
    """
    if message_type is None:
        message_type = "comment"
    try:
        category = _MessageCategoryMapping[message_type.lower()]
    except KeyError:
        raise ValueError(
            f"'{message_type}' is not a valid e-logbook message type"
        ) from None

    if category != _MessageCategory.comment:
        # Non-comments cannot be editable
        if editable:
            warnings.warn(
                f"message type '{message_type}' cannot be editable", UserWarning
            )
            editable = None

        # Non-comments cannot be unformatted
        if formatted is not None and not formatted:
            warnings.warn(
                f"message type '{message_type}' cannot be unformatted", UserWarning
            )
            formatted = None

    # A message cannot be editable and formatted at the same time or uneditable and unformatted
    if formatted == editable and formatted is not None:
        if formatted:
            warnings.warn(
                f"message type '{message_type}' cannot be editable and formatted at the same time",
                UserWarning,
            )
        else:
            warnings.warn(
                f"message type '{message_type}' cannot be uneditable and unformatted at the same time",
                UserWarning,
            )
        formatted = None
        editable = None

    # Editability is specified
    if editable is not None:
        if editable:
            return category, _MessageType.annotation
        return category, _MessageType.notification

    # Formatability is specified
    if formatted is not None:
        if formatted:
            return category, _MessageType.notification
        return category, _MessageType.annotation

    # By default comments are annotations and the rest are notifications
    if category == _MessageCategory.comment:
        return category, _MessageType.annotation
    return category, _MessageType.notification



import random
import datetime
from typing import Optional

import pytest

from ..client.investigation import _select_investigation


def test_investigation_none():
    investigations = _Investigations()

    # No investigations
    investigations.check(datetime.timedelta())

    # Ignore open-ended
    # _____|‾‾‾‾‾
    #         |
    investigations.session(datetime.timedelta(days=1, seconds=1))
    investigations.check(datetime.timedelta(days=2), allow_open_ended=False)


def test_investigation_inside():
    investigations = _Investigations()

    # Inside single investigation
    # _____|‾‾‾‾‾|_____
    #        |
    investigations.session(
        datetime.timedelta(days=1), datetime.timedelta(days=3), expected=True
    )
    investigations.check(datetime.timedelta(days=1, seconds=1))

    # Inside the earliest of two investigation
    # _____|‾‾‾‾‾|___________
    # ___________|‾‾‾‾‾|_____
    #        |
    investigations.session(
        datetime.timedelta(days=1), datetime.timedelta(days=2), expected=True
    )
    investigations.session(datetime.timedelta(days=2), datetime.timedelta(days=3))
    investigations.check(datetime.timedelta(days=1, hours=2))

    # ___________|‾‾‾‾‾|_____
    # _____|‾‾‾‾‾|___________
    #        |
    investigations.session(datetime.timedelta(days=2), datetime.timedelta(days=3))
    investigations.session(
        datetime.timedelta(days=1), datetime.timedelta(days=2), expected=True
    )
    investigations.check(datetime.timedelta(days=1, hours=2))

    # On the border of two subsequent investigations
    # _____|‾‾‾‾‾|___________
    # ___________|‾‾‾‾‾|_____
    #            |
    investigations.session(datetime.timedelta(days=1), datetime.timedelta(days=2))
    investigations.session(
        datetime.timedelta(days=2), datetime.timedelta(days=3), expected=True
    )
    investigations.check(datetime.timedelta(days=2))

    # ___________|‾‾‾‾‾|_____
    # _____|‾‾‾‾‾|___________
    #            |
    investigations.session(datetime.timedelta(days=2), datetime.timedelta(days=3))
    investigations.session(
        datetime.timedelta(days=1), datetime.timedelta(days=2), expected=True
    )
    investigations.check(datetime.timedelta(days=2))

    # Inside closed-ended and open-ended investigation
    # _____|‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
    # ___________|‾‾‾‾‾|_____
    #              |
    investigations.session(datetime.timedelta(days=1))
    investigations.session(
        datetime.timedelta(days=2), datetime.timedelta(days=3), expected=True
    )
    investigations.check(datetime.timedelta(days=2, seconds=1))

    # ___________|‾‾‾‾‾|_____
    # _____|‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
    #              |
    investigations.session(
        datetime.timedelta(days=2), datetime.timedelta(days=3), expected=True
    )
    investigations.session(datetime.timedelta(days=1))
    investigations.check(datetime.timedelta(days=2, seconds=1))

    # On the border of closed-ended and open-ended investigation (not same start date)
    # _____|‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
    # ___________|‾‾‾‾‾|_____
    #            |
    investigations.session(datetime.timedelta(days=1))
    investigations.session(
        datetime.timedelta(days=2), datetime.timedelta(days=3), expected=True
    )
    investigations.check(datetime.timedelta(days=2))

    # ___________|‾‾‾‾‾|_____
    # _____|‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
    #            |
    investigations.session(
        datetime.timedelta(days=2), datetime.timedelta(days=3), expected=True
    )
    investigations.session(datetime.timedelta(days=1))
    investigations.check(datetime.timedelta(days=2))

    # On the border of closed-ended and open-ended investigation (same start date)
    # ___________|‾‾‾‾‾‾‾‾‾‾‾
    # ___________|‾‾‾‾‾|_____
    #            |
    investigations.session(datetime.timedelta(days=2))
    investigations.session(
        datetime.timedelta(days=2), datetime.timedelta(days=3), expected=True
    )
    investigations.check(datetime.timedelta(days=2))

    investigations.session(datetime.timedelta(days=2))
    investigations.session(
        datetime.timedelta(days=2), datetime.timedelta(days=3), expected=True
    )
    investigations.check(datetime.timedelta(days=2), allow_open_ended=False)

    # ___________|‾‾‾‾‾|_____
    # ___________|‾‾‾‾‾‾‾‾‾‾‾
    #            |
    investigations.session(
        datetime.timedelta(days=2), datetime.timedelta(days=3), expected=True
    )
    investigations.session(datetime.timedelta(days=2))
    investigations.check(datetime.timedelta(days=2))

    investigations.session(
        datetime.timedelta(days=2), datetime.timedelta(days=3), expected=True
    )
    investigations.session(datetime.timedelta(days=2))
    investigations.check(datetime.timedelta(days=2), allow_open_ended=False)

    # Select inside overlapping open-ended investigations
    # _____|‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
    # ___________|‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
    # _________________|‾‾‾‾‾‾‾‾‾‾‾
    #        |
    investigations.session(datetime.timedelta(days=1), expected=True)
    investigations.session(datetime.timedelta(days=2))
    investigations.session(datetime.timedelta(days=3))
    investigations.check(datetime.timedelta(days=1, seconds=1))

    # _____|‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
    # ___________|‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
    # _________________|‾‾‾‾‾‾‾‾‾‾‾
    #             |
    investigations.session(datetime.timedelta(days=1))
    investigations.session(datetime.timedelta(days=2), expected=True)
    investigations.session(datetime.timedelta(days=3))
    investigations.check(datetime.timedelta(days=2, seconds=1))

    # _____|‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
    # ___________|‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
    # _________________|‾‾‾‾‾‾‾‾‾‾‾
    #                   |
    investigations.session(datetime.timedelta(days=1))
    investigations.session(datetime.timedelta(days=2))
    investigations.session(datetime.timedelta(days=3), expected=True)
    investigations.check(datetime.timedelta(days=3, seconds=1))

    # Inside two overlapping closed-ended investigations
    # _____|‾‾‾‾‾|________
    # ________|‾‾‾‾‾|_____
    #           |
    investigations.session(datetime.timedelta(days=1), datetime.timedelta(days=3))
    investigations.session(
        datetime.timedelta(days=2), datetime.timedelta(days=4), expected=True
    )
    investigations.check(datetime.timedelta(days=2, hours=2))

    # ________|‾‾‾‾‾|_____
    # _____|‾‾‾‾‾|________
    #           |
    investigations.session(datetime.timedelta(days=2), datetime.timedelta(days=4))
    investigations.session(
        datetime.timedelta(days=1), datetime.timedelta(days=3), expected=True
    )
    investigations.check(datetime.timedelta(days=2, hours=2))

    # Inside two overlapping open and closed-ended investigations
    # _____|‾‾‾‾‾‾‾‾‾‾‾‾‾‾
    # ________|‾‾‾‾‾|_____
    #           |
    investigations.session(datetime.timedelta(days=1))
    investigations.session(
        datetime.timedelta(days=2), datetime.timedelta(days=4), expected=True
    )
    investigations.check(datetime.timedelta(days=2, hours=2))

    # ________|‾‾‾‾‾|_____
    # _____|‾‾‾‾‾‾‾‾‾‾‾‾‾‾
    #           |
    investigations.session(
        datetime.timedelta(days=2), datetime.timedelta(days=4), expected=True
    )
    investigations.session(datetime.timedelta(days=1))
    investigations.check(datetime.timedelta(days=2, hours=2))

    # _____|‾‾‾‾‾‾‾‾‾‾‾‾‾‾
    # ________|‾‾‾‾‾|_____
    #         |
    investigations.session(datetime.timedelta(days=1))
    investigations.session(
        datetime.timedelta(days=2), datetime.timedelta(days=4), expected=True
    )
    investigations.check(datetime.timedelta(days=2))

    # ________|‾‾‾‾‾|_____
    # _____|‾‾‾‾‾‾‾‾‾‾‾‾‾‾
    #         |
    investigations.session(
        datetime.timedelta(days=2), datetime.timedelta(days=4), expected=True
    )
    investigations.session(datetime.timedelta(days=1))
    investigations.check(datetime.timedelta(days=2))


def test_investigation_inbetween():
    investigations = _Investigations()

    # Between two investigations
    # _____|‾‾‾‾‾|_________________
    # _________________|‾‾‾‾‾|_____
    #             |
    investigations.session(
        datetime.timedelta(days=1), datetime.timedelta(days=2), expected=True
    )
    investigations.session(datetime.timedelta(days=3), datetime.timedelta(days=4))
    investigations.check(datetime.timedelta(days=2, hours=2))

    # _____|‾‾‾‾‾|_________________
    # _________________|‾‾‾‾‾|_____
    #                 |
    investigations.session(datetime.timedelta(days=1), datetime.timedelta(days=2))
    investigations.session(
        datetime.timedelta(days=3), datetime.timedelta(days=4), expected=True
    )
    investigations.check(datetime.timedelta(days=2, hours=13))

    # _____|‾‾‾‾‾|_________________
    # _________________|‾‾‾‾‾‾‾‾‾‾‾
    #             |
    investigations.session(
        datetime.timedelta(days=1), datetime.timedelta(days=2), expected=True
    )
    investigations.session(datetime.timedelta(days=3))
    investigations.check(datetime.timedelta(days=2, hours=2))

    # _____|‾‾‾‾‾|_________________
    # _________________|‾‾‾‾‾‾‾‾‾‾‾
    #                 |
    investigations.session(datetime.timedelta(days=1), datetime.timedelta(days=2))
    investigations.session(datetime.timedelta(days=3), expected=True)
    investigations.check(datetime.timedelta(days=2, hours=13))

    # _____|‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
    # __________|‾‾‾‾‾|_________________
    # ______________________|‾‾‾‾‾|_____
    #                      |
    investigations.session(datetime.timedelta(days=1))
    investigations.session(datetime.timedelta(days=2), datetime.timedelta(days=3))
    investigations.session(
        datetime.timedelta(days=5), datetime.timedelta(days=6), expected=True
    )
    investigations.check(datetime.timedelta(days=4, hours=13))

    # __________|‾‾‾‾‾|_________________
    # ______________________|‾‾‾‾‾|_____
    # _____|‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
    #                      |
    investigations.session(datetime.timedelta(days=2), datetime.timedelta(days=3))
    investigations.session(
        datetime.timedelta(days=5), datetime.timedelta(days=6), expected=True
    )
    investigations.session(datetime.timedelta(days=1))
    investigations.check(datetime.timedelta(days=4, hours=13))


def test_investigation_outside():
    investigations = _Investigations()

    # Date before investigation
    # _____|‾‾‾‾‾|_____
    #    |
    investigations.session(
        datetime.timedelta(days=1, seconds=1), datetime.timedelta(days=2), expected=True
    )
    investigations.check(datetime.timedelta(days=1))

    # Date before open-ended investigation
    # _____|‾‾‾‾‾
    # |
    investigations.session(datetime.timedelta(days=1), expected=True)
    investigations.check(datetime.timedelta())

    # Date after investigation
    # _____|‾‾‾‾‾|_____
    #             |
    investigations.session(
        datetime.timedelta(days=1), datetime.timedelta(days=2), expected=True
    )
    investigations.check(datetime.timedelta(days=2, seconds=1))

    # Date before two investigations
    # _____|‾‾‾‾‾|_________________
    # _________________|‾‾‾‾‾|_____
    #  |
    investigations.session(
        datetime.timedelta(days=1), datetime.timedelta(days=2), expected=True
    )
    investigations.session(datetime.timedelta(days=3), datetime.timedelta(days=4))
    investigations.check(datetime.timedelta(hours=2))

    # Date after two investigations
    # _____|‾‾‾‾‾|_________________
    # _________________|‾‾‾‾‾|_____
    #                          |
    investigations.session(datetime.timedelta(days=1), datetime.timedelta(days=2))
    investigations.session(
        datetime.timedelta(days=3), datetime.timedelta(days=4), expected=True
    )
    investigations.check(datetime.timedelta(days=4, hours=2))


def test_investigation_dateonly():
    investigations = _Investigations()

    # One investigation starts on the same day
    # _____|‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
    # ___________|‾‾‾‾‾|_____
    #           |
    investigations.session(datetime.timedelta(days=1))
    investigations.session(
        datetime.timedelta(days=3), datetime.timedelta(days=4), expected=True
    )
    with pytest.raises(AssertionError):
        investigations.check(datetime.timedelta(days=2, hours=23))

    investigations.session(datetime.timedelta(days=1))
    investigations.session(
        datetime.timedelta(days=3), datetime.timedelta(days=4), expected=True
    )
    investigations.check(datetime.timedelta(days=2, hours=23), date_only=True)

    # Two investigations start on the same day
    # _____|‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾
    # ___________|‾‾‾‾‾|_____
    # ___________|‾‾‾‾‾|_____
    #           |
    investigations.session(datetime.timedelta(days=1))
    investigations.session(datetime.timedelta(days=3), datetime.timedelta(days=4))
    investigations.session(
        datetime.timedelta(days=3), datetime.timedelta(days=4), expected=True
    )
    investigations.check(datetime.timedelta(days=2, hours=23), date_only=True)


class _Investigations:
    def __init__(self):
        self._reset()

    def session(
        self,
        start_offset: datetime.timedelta,
        end_offset: Optional[datetime.timedelta] = None,
        expected: bool = False,
    ) -> None:
        startdate = self._date + start_offset
        investigation = {
            "id": len(self._investigations),
            "startDate": startdate.astimezone().isoformat(),
            "unique": random.uniform(0, 1),
        }
        if end_offset is not None:
            enddate = self._date + end_offset
            investigation["endDate"] = enddate.astimezone().isoformat()
        self._investigations.append(investigation)
        if expected:
            self._expected = investigation

    def _reset(self):
        self._investigations = list()
        self._expected = None
        self._nextid = 10000
        # +1 day: CEST -> EST
        self._date = datetime.datetime(year=2023, month=10, day=28, hour=8)

    def check(
        self,
        offset: datetime.timedelta,
        allow_open_ended: bool = True,
        date_only: bool = False,
    ):
        if self._investigations:
            random.shuffle(self._investigations)
        date = (self._date + offset).astimezone()
        if date_only:
            date = date.date()
        investigation = _select_investigation(
            self._investigations, date=date, allow_open_ended=allow_open_ended
        )
        assert investigation == self._expected
        self._reset()



import os
from typing import Optional

from blissdata.beacon.files import read_config

from .main import IcatClient


def get_icat_client(
    timeout: Optional[float] = None,
) -> IcatClient:
    beacon_host = os.environ.setdefault("BEACON_HOST", "id00:25000")

    url = f"beacon://{beacon_host}/__init__.yml"
    config = read_config(url).get("icat_servers")

    if not config:
        raise RuntimeError(
            f"Beacon host {beacon_host} does not provide ICAT configuration"
        )
    return IcatClient(
        metadata_urls=config["metadata_urls"],
        elogbook_url=config["elogbook_url"],
        elogbook_token=config["elogbook_token"],
        feedback_timeout=timeout,
        add_files_urls=config["metadata_urls"],
    )



from requests.exceptions import HTTPError

import pytest

from .utils import generate


def test_login(icatplus_restricted_client):
    client, messages = icatplus_restricted_client

    with pytest.raises(RuntimeError, match="Login is required."):
        _ = client.session_id

    with pytest.raises(HTTPError, match="403 Client Error: Authentication failed"):
        _ = client.login("wrong")

    result = client.login("correct")

    assert result["sessionId"] == client.session_id

    assert messages.empty()


def test_get_investigations_by(icatplus_restricted_client, icat_metadata_client):
    client, messages = icatplus_restricted_client
    mclient, mmessages = icat_metadata_client

    with pytest.raises(RuntimeError, match="Login is required."):
        _ = client.get_investigations_by()

    assert client.login("correct")

    investigations = client.get_investigations_by()
    assert isinstance(investigations, list)

    mclient.start_investigation(proposal="hg123", beamline="id00")
    message = mmessages.get(timeout=10)

    investigations = client.get_investigations_by()
    expected = [
        {
            "experiment": "hg123",
            "id": 0,
            "instrument": {
                "name": "id00",
            },
            "proposal": "hg123",
            "startDate": message["investigation"]["startDate"],
        }
    ]
    assert investigations == expected

    assert messages.empty()
    assert mmessages.empty()


def test_get_parcels_by(icatplus_restricted_client):
    client, messages = icatplus_restricted_client

    investigation_id = generate.investigation_id()

    with pytest.raises(RuntimeError, match="Login is required."):
        _ = client.get_parcels_by(investigation_id)

    assert client.login("correct")

    parcels = client.get_parcels_by(investigation_id)

    assert parcels == []
    assert messages.empty()



import os
import json
import logging
import datetime
from tqdm import tqdm
from glob import glob
import concurrent.futures
from typing import Dict, Any, List, Tuple, Optional, Callable

from . import raw_data
from . import sync_types
from . import path_utils

logger = logging.getLogger("ICAT SYNC")


class ExperimentalSessionStore:
    def __init__(
        self,
        cache_dir: Optional[str] = None,
        save_dir: Optional[str] = None,
        raw_data_format: str = "esrfv3",
        invalidate_cache: bool = False,
    ) -> None:
        """
        :param cache_dir: one JSON file for each cached experimental session
        :param save_dir: one CSV+BASH file for each sync status
        :param raw_data_format: will be used as subdirectory name
        :param invalidate_cache: invalidating session when loaded from cache
        """
        self._save_dir = save_dir
        self._cache_dir = cache_dir
        self._raw_data_format = raw_data_format
        self._invalidate_cache = invalidate_cache

        self._save_subdir = None
        self._cache_subdir = None
        if save_dir is not None:
            self._save_subdir = os.path.join(save_dir, raw_data_format)
        if cache_dir is not None:
            self._cache_subdir = os.path.join(cache_dir, raw_data_format)

        self._exp_sessions: Dict[str, sync_types.ExperimentalSession] = dict()
        self._load_all_session_files()
        self._start_time = datetime.datetime.now()

    def __enter__(self) -> "ExperimentalSessionStore":
        self._start_time = datetime.datetime.now()
        return self

    def __exit__(self, *_) -> bool:
        self.close()
        return False

    def get_session(self, session_dir: str) -> Optional[sync_types.ExperimentalSession]:
        return self._exp_sessions.get(session_dir)

    def add_session(self, exp_session: sync_types.ExperimentalSession) -> None:
        """Add session to the in-memory cache and when enabled save it on disk."""
        self._exp_sessions[exp_session.session_dir] = exp_session
        if not self._cache_subdir:
            return
        filename = self._cache_filename(exp_session)
        data = exp_session.as_dict()
        os.makedirs(self._cache_subdir, mode=0o775, exist_ok=True)
        _json_save(data, filename)

    def remove_session(self, exp_session: sync_types.ExperimentalSession) -> None:
        """Remove session from the in-memory cache and when enabled from disk."""
        self._exp_sessions.pop(exp_session.session_dir, None)
        if not self._cache_subdir:
            return
        filename = self._cache_filename(exp_session)
        try:
            os.remove(filename)
        except FileNotFoundError:
            pass
        logger.info("Removed session cache %s", filename)

    def close(self) -> None:
        try:
            self._save_overview()
        finally:
            self._print_final()

    def _load_all_session_files(self) -> None:
        """Load all the session files when enabled and delete the ones for sessions which directory no longer exists."""
        if not self._cache_subdir:
            return
        logger.info("Loading cache from %s ...", self._cache_subdir)
        filenames = glob(os.path.join(self._cache_subdir, "*.json"))

        logger.info("Loading cache of %d sessions ...", len(filenames))
        data = _executor_map_with_progress(_json_load, filenames)

        if self._invalidate_cache:
            logger.info("Invalidate cache of %d sessions ...", len(data))

        if self._invalidate_cache:
            func = _parse_session_cache_with_invalidation
        else:
            func = _parse_session_cache
        exp_sessions = _executor_map_with_progress(func, filenames, data)

        for exp_session in exp_sessions:
            if exp_session is not None:
                self._exp_sessions[exp_session.session_dir] = exp_session

        logger.info("Loaded cache of %d sessions.", len(self._exp_sessions))

    def _cache_filename(self, exp_session: sync_types.ExperimentalSession) -> str:
        basename = (
            f"{exp_session.proposal}_{exp_session.beamline}_{exp_session.session}.json"
        )
        return os.path.join(self._cache_subdir, basename)

    def _save_overview(self) -> None:
        if not self._save_subdir:
            return
        logger.info(
            "Saving synchronization overview of %d sessions ...",
            len(self._exp_sessions),
        )
        os.makedirs(self._save_subdir, mode=0o775, exist_ok=True)

        save_columns = [
            "beamline",
            "proposal",
            "timeslot",
            "session",
            "url",
            "search_url",
            "path",
            *sync_types.ExperimentalSession.DATASET_STATUSES,
        ]

        save_category = [
            "todo",
            "ok",
            "empty",
            "has_invalid",
            "todo_ongoing",
            "not_in_icat",
        ]

        register_cmd_params = ""
        if self._save_dir:
            register_cmd_params += f"--save-dir {self._save_dir} "
        if self._cache_dir:
            register_cmd_params += f"--cache-dir {self._cache_dir} "
        register_cmd_params += '"$@"'

        with concurrent.futures.ThreadPoolExecutor() as executor:
            # Clear CSV files and add header
            stats_files = dict(
                (s, os.path.join(self._save_subdir, s + ".csv")) for s in save_category
            )

            def _init_csv(filename):
                _save_csv_line(filename, save_columns, clear=True)
                try:
                    os.chmod(filename, 0o0664)
                except PermissionError:
                    pass

            _ = list(executor.map(_init_csv, stats_files.values()))

            # Clear bash files and add header
            cmd_files = dict(
                (s, os.path.join(self._save_subdir, s + ".sh")) for s in save_category
            )

            def _init_cmd(filename):
                _save_file_line(filename, "#!/bin/bash", clear=True)
                _save_file_line(filename, "")
                _save_file_line(
                    filename,
                    "# Run a cached copy because this script can be modified during execution",
                )
                _save_file_line(filename, 'if [ -z "$SCRIPT_CACHED" ]; then')
                _save_file_line(filename, '    SCRIPT_CACHED=$(<"$0")')
                _save_file_line(filename, '    script_name=$(basename "$0")')
                _save_file_line(
                    filename, '    echo "$SCRIPT_CACHED" > /tmp/cached_$script_name'
                )
                _save_file_line(
                    filename,
                    '    SCRIPT_CACHED=1 /bin/bash /tmp/cached_$script_name "$@"',
                )
                _save_file_line(filename, "    exit $?")
                _save_file_line(filename, "fi")
                _save_file_line(filename, "")
                try:
                    os.chmod(filename, 0o0774)
                except PermissionError:
                    pass

            _ = list(executor.map(_init_cmd, cmd_files.values()))

            # Add sessions CSV and BASH files
            def _save_session_line(exp_session):
                icat_investigation = exp_session.icat_investigation
                if not icat_investigation:
                    save_category = "not_in_icat"
                else:
                    statuses = exp_session.dataset_statuses()
                    if statuses & {
                        "not_uploaded",
                        "unregistered",
                        "registered_without_files",
                    }:
                        if icat_investigation and icat_investigation.ongoing:
                            save_category = "todo_ongoing"
                        else:
                            save_category = "todo"
                    elif "invalid" in statuses:
                        save_category = "has_invalid"
                    elif "registered" in statuses:
                        save_category = "ok"
                    else:
                        save_category = "empty"

                tabular_data = exp_session.tabular_data()
                stats = [tabular_data[k] for k in save_columns]
                _save_csv_line(stats_files[save_category], stats)

                register_cmd = f"icat-sync-raw --beamline {exp_session.beamline} --proposal {exp_session.proposal} --session {exp_session.session} {register_cmd_params}"
                _save_file_line(cmd_files[save_category], register_cmd)

            _ = list(executor.map(_save_session_line, self._exp_sessions.values()))

        logger.info("Saving synchronization finished.")

    def _print_final(self):
        end_time = datetime.datetime.now()
        logger.info("")
        logger.info("Total synchronization time was %s", end_time - self._start_time)
        if self._cache_subdir:
            logger.info("Cache was used from %s", self._cache_subdir)
        if self._save_subdir:
            logger.info("Results can be found in %s", self._save_subdir)
            logger.info(
                "To fix issues run %s", os.path.join(self._save_subdir, "todo.sh")
            )


def _executor_map_with_progress(func: Callable, *args) -> List[Any]:
    results = []
    with concurrent.futures.ThreadPoolExecutor() as executor:
        with tqdm(total=len(args[0])) as progress:

            def update_progress(future):
                progress.update(1)
                results.append(future.result())

            futures = [executor.submit(func, *_args) for _args in zip(*args)]

            for future in futures:
                future.add_done_callback(update_progress)

            result = concurrent.futures.wait(futures)

            for future in result.done:
                if future.exception():
                    raise future.exception()

    return results


def _parse_session_cache(
    filename: str, data: dict
) -> Optional[sync_types.ExperimentalSession]:
    try:
        return sync_types.ExperimentalSession.from_dict(data)
    except Exception as e:
        logger.info("Invalidate %s: wrong cache format (%s)", filename, e)
        try:
            os.remove(filename)
        except FileNotFoundError:
            pass


def _parse_session_cache_with_invalidation(
    filename: str, data: dict
) -> Optional[sync_types.ExperimentalSession]:
    exp_session = _parse_session_cache(filename, data)
    if exp_session is None:
        return

    invalidation_reason = _session_invalidation_reason(exp_session)
    if invalidation_reason:
        logger.info("Invalidate %s: %s", exp_session.session_dir, invalidation_reason)
        try:
            os.remove(filename)
        except FileNotFoundError:
            pass
        return

    return exp_session


def _session_invalidation_reason(
    exp_session: sync_types.ExperimentalSession,
) -> Optional[str]:
    if not os.path.isdir(exp_session.session_dir):
        return "no longer exists"

    if not exp_session.icat_investigation:
        # Cached session does not contain any datasets when
        # no ICAT investigation was found
        return

    cached = {path_utils.markdir(dset.path) for dset in exp_session.iter_datasets()}
    dataset_filters = raw_data.get_dataset_filters(
        exp_session.raw_root_dir, raw_data_format=exp_session.raw_data_format
    )
    actual = {
        path_utils.markdir(path)
        for dataset_filter in dataset_filters
        for path in glob(path_utils.markdir(dataset_filter))
    }
    if cached != actual:
        nnew = len(actual - cached)
        nremoved = len(cached - actual)
        return f"has {nnew} new datasets and {nremoved} datasets were removed"


def _json_save(data: Dict[str, Any], filename: str) -> None:
    try:
        with open(filename, "w") as fp:
            json.dump(data, fp, default=_json_serializer, indent=2)
    except Exception as e:
        raise IOError(f"writing failed for file {filename}") from e


def _json_load(filename: str) -> Dict[str, Any]:
    try:
        with open(filename, "r") as fp:
            return json.load(fp, object_pairs_hook=_json_pair_deserializer)
    except Exception as e:
        raise IOError(f"reading failed for file {filename}") from e


def _json_serializer(obj: Any):
    """Serialize datetime and date objects to string."""
    if isinstance(obj, (datetime.datetime, datetime.date)):
        return obj.isoformat()
    else:
        raise TypeError(
            "Object of type '{}' is not JSON serializable".format(type(obj))
        )


def _json_pair_deserializer(items: List[Tuple[str, Any]]) -> Dict[str, Any]:
    """Deserialize string to datetime or date objects."""
    result = dict()
    for key, value in items:
        if "date" in key and value:
            try:
                value = datetime.date.fromisoformat(value)
            except ValueError:
                value = datetime.datetime.fromisoformat(value)
        result[key] = value
    return result


def _save_csv_line(filename: str, row: List[Any], clear: bool = False) -> None:
    line = ",".join([str(item).replace(",", " ").strip() for item in row])
    _save_file_line(filename, line, clear)


def _save_file_line(filename: str, line: str, clear: bool = False) -> None:
    mode = "w" if clear else "a"
    with open(filename, mode) as f:
        f.write(f"{line}\n")



import os
import json
import socket
import logging
import threading
from enum import Enum
from typing import Optional
from json.decoder import JSONDecodeError

import stomp
import xmltodict
from xml.parsers.expat import ExpatError

from .icat_db import IcatDb
from ...utils.log_utils import basic_config
from .activemq_rest_server import ICAT_QUEUES

logger = logging.getLogger("STOMP SUBSCRIBER")

MessageType = Enum("MessageType", "investigation dataset archiving addfiles unknown")


class MyListener(stomp.ConnectionListener):
    def __init__(self, conn, icat_data_dir: Optional[str] = None):
        self.conn = conn
        self.s_out = None
        self.icatdb = IcatDb(icat_data_dir)
        super().__init__()

    def redirect_messages(self, port):
        if self.s_out is not None:
            self.s_out.close()
        self.s_out = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.s_out.connect(("localhost", port))
        logger.info(f"Redirect received messages to port {port}")

    def on_message(self, frame):
        message = frame.body
        logger.info("received message:\n %s", message)

        message_type = None
        try:
            message = xmltodict.parse(
                message,
                process_namespaces=True,
                namespaces={"http://www.esrf.fr/icat": None},
            )
        except ExpatError:
            try:
                message = json.loads(message)
            except JSONDecodeError:
                message_type = MessageType.unknown
            else:
                keys = set(message)
                if keys == {"datasetId", "type", "level", "message"}:
                    message_type = MessageType.archiving
                elif keys == {"datasetId"}:
                    message_type = MessageType.addfiles
                else:
                    message_type = MessageType.unknown
        else:
            if "investigation" in message:
                message_type = MessageType.investigation
            elif "dataset" in message:
                message_type = MessageType.dataset

        logger.info("parsed message (%s):\n %s", message_type, message)

        # Only access specific destinations
        header = frame.headers
        icat_queues = ["/queue/" + q for q in ICAT_QUEUES]
        if header.get("destination") not in icat_queues:
            return

        # Only accept valid proposals
        if message_type in (message_type.investigation, message_type.dataset):
            if message_type is message_type.investigation:
                data = message["investigation"]
                proposal = data["experiment"]
            else:
                data = message["dataset"]
                proposal = data["investigation"]
                # Convert to backend format
                file_count = 0
                if os.path.exists(data["location"]):
                    file_count = len(os.listdir(data["location"]))
                data["parameter"].append({"name": "__fileCount", "value": file_count})
                data["parameters"] = data.pop("parameter")
            if proposal and "666" in proposal:
                logger.info(
                    "Do not register %s for invalid proposal '%s'",
                    message_type,
                    proposal,
                )
                return

        # Store data
        if message_type in (message_type.investigation, message_type.dataset):
            if message_type is message_type.investigation:
                self.icatdb.start_investigation(data)
            else:
                self.icatdb.store_dataset(data)

        if message_type == message_type.addfiles:
            dataset_id = message["datasetId"]
            with self.icatdb.update_dataset(dataset_id) as data:
                if data is None:
                    logger.error("datasetId %s does not exist", dataset_id)
                else:
                    file_count = 0
                    dirname = data["location"]
                    if os.path.exists(dirname):
                        file_count = len(os.listdir(dirname))
                    for parameter in data["parameters"]:
                        if parameter["name"] == "__fileCount":
                            logger.info(
                                "Update file count for %s to %d", dirname, file_count
                            )
                            parameter["value"] = file_count
                            break
                    else:
                        logger.info("Add file count for %s to %d", dirname, file_count)
                        data["parameters"].append(
                            {"name": "__fileCount", "value": file_count}
                        )

        # Notify that data is valid
        if self.s_out is not None:
            self.s_out.sendall(frame.body.encode() + b"\n")


def main(
    host=None, port=60001, queue=None, port_out=0, icat_data_dir: Optional[str] = None
):
    if not host:
        host = "localhost"
    if not queue:
        queue = "/queue/icatIngest"
    conn = stomp.Connection([(host, port)])
    # Listener will run in a different thread
    listener = MyListener(conn, icat_data_dir)
    conn.set_listener("", listener)
    conn.connect("guest", "guest", wait=True)
    conn.subscribe(destination=queue, id=1, ack="auto")
    logger.info(f"subscribed to {queue} on STOMP {host}:{port}")
    if port_out:
        listener.redirect_messages(port_out)
        listener.s_out.sendall(b"LISTENING\n")
    logger.info("CTRL-C to stop")
    try:
        threading.Event().wait()
    finally:
        logger.info("Exit.")


if __name__ == "__main__":
    import argparse

    basic_config(
        logger=logger,
        level=logging.DEBUG,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )

    parser = argparse.ArgumentParser(
        description="STOMP client which subscribes to a STOMP queue and redirect its output to a socket"
    )
    parser.add_argument(
        "--host", default="localhost", type=str, help="STOMP server host"
    )
    parser.add_argument("--port", default=60001, type=int, help="STOMP server port")
    parser.add_argument(
        "--queue", default="/queue/icatIngest", type=str, help="STOMP queue"
    )
    parser.add_argument("--port_out", default=0, type=int, help="output socket")
    parser.add_argument(
        "--icat_data_dir", default=None, type=str, help="Dataset directory"
    )
    args = parser.parse_args()

    main(
        host=args.host,
        port=args.port,
        port_out=args.port_out,
        queue=args.queue,
        icat_data_dir=args.icat_data_dir,
    )



import json
from typing import Optional, List
from .messaging import IcatMessagingClient
from . import defaults


class IcatAddFilesClient:
    """Client for adding missing files to an existing dataset"""

    def __init__(
        self,
        queue_urls: Optional[List[str]] = None,
        queue_name: Optional[str] = None,
        monitor_port: Optional[int] = None,
        timeout: Optional[float] = None,
    ):
        if queue_name is None:
            queue_name = defaults.ADD_FILES_QUEUE
        if queue_urls is None:
            queue_urls = defaults.ADD_FILES_BROKERS
        self._client = IcatMessagingClient(
            queue_urls, queue_name, monitor_port=monitor_port, timeout=timeout
        )

    def disconnect(self):
        self._client.disconnect()

    def add_files(
        self,
        dataset_id: int,
    ):
        assert dataset_id, "ICAT requires the datasetId"
        root = {"datasetId": dataset_id}
        data = json.dumps(root).encode("utf-8")
        self._client.send(data)

    def check_health(self):
        """Raises an exception when not healthy"""
        self._client.check_health()



import os
import json
import logging
import datetime
from glob import glob
from contextlib import contextmanager
from typing import Optional, Dict, List, Generator


logger = logging.getLogger("STOMP SUBSCRIBER")


class IcatDb:
    def __init__(self, root_dir: Optional[str] = None):
        if root_dir is None:
            root_dir = "."
        if root_dir:
            os.makedirs(root_dir, exist_ok=True)
        self._root_dir = root_dir
        self._investigations = os.path.join(root_dir, "investigations.json")

    def start_investigation(self, investigation: dict) -> None:
        investigation["instrument"] = {"name": investigation["instrument"]}
        investigation["proposal"] = investigation["experiment"]
        _add_table(self._investigations, investigation)

    def store_dataset(self, dataset: dict) -> None:
        investigation = _find_data(self._investigations, dataset["startDate"])
        if investigation is None:
            logger.error(
                "Dataset not stored because no investigation found: %s", dataset
            )
            return
        filename = os.path.join(self._root_dir, f"datasets{investigation['id']}.json")
        dataset["id"] = self._get_next_dataset_id()
        _add_table(filename, dataset)

    @contextmanager
    def update_dataset(self, dataset_id: int) -> Generator[dict, None, None]:
        for filename in glob(os.path.join(self._root_dir, "datasets*.json")):
            data = _get_row(filename, dataset_id)
            if data is None:
                continue
            yield data
            _edit_table(filename, data)
            break
        else:
            yield

    def get_all_investigations(self) -> List[Dict]:
        return _read_table(self._investigations)

    def get_investigations(self, instrument: str, experiment: str) -> List[Dict]:
        return [
            investigation
            for investigation in _read_table(self._investigations)
            if experiment == investigation["experiment"]
            and instrument == investigation["instrument"]["name"]
        ]

    def get_datasets(self, investigation_id) -> List[Dict]:
        filename = os.path.join(self._root_dir, f"datasets{investigation_id}.json")
        return _read_table(filename)

    def _get_next_dataset_id(self) -> int:
        max_ids = [
            _get_max_id(filename)
            for filename in glob(os.path.join(self._root_dir, "datasets*.json"))
        ]
        if max_ids:
            return max(max_ids) + 1
        return 0


def _read_table(filename: str) -> List[dict]:
    if not os.path.isfile(filename):
        return list()
    with open(filename, "r") as f:
        return json.load(f)


def _write_table(filename: str, data: dict) -> Dict:
    with open(filename, "w") as f:
        json.dump(data, f)
    logger.info("Metadata in %s updated", filename)


def _add_table(filename: str, data: dict) -> None:
    table = _read_table(filename)
    if "id" not in data:
        if table:
            data_id = max(row["id"] for row in table) + 1
        else:
            data_id = 0
        data = dict(data)
        data["id"] = data_id
    table.append(data)
    logger.info("Add metadata to %s: %s", filename, data)
    _write_table(filename, table)


def _edit_table(filename: str, data: dict) -> None:
    table = _read_table(filename)
    for row in table:
        if row["id"] == data["id"]:
            logger.info("Edit metadata of %s: %s", filename, data)
            row.update(data)
            _write_table(filename, table)
            break
    else:
        _add_table(filename, data)


def _get_row(filename: str, dataset_id: int) -> Optional[dict]:
    table = _read_table(filename)
    for row in table:
        if row["id"] == dataset_id:
            return row


def _get_max_id(filename: str) -> int:
    table = _read_table(filename)
    if table:
        return max(row["id"] for row in table)
    return -1


def _find_data(filename: str, date: str) -> Optional[Dict]:
    date = datetime.datetime.fromisoformat(date).astimezone()
    for data in _read_table(filename):
        start_date = datetime.datetime.fromisoformat(data["startDate"]).astimezone()
        end_date = data.get("endDate")
        if end_date is None:
            # infinite timeslot
            inside_timeslot = date >= start_date
        else:
            # finite timeslot
            end_date = datetime.datetime.fromisoformat(end_date).astimezone()
            inside_timeslot = date >= start_date and date <= end_date
        if inside_timeslot:
            return data
